import numpy as np
from .Utilities import createLog
from .Utilities import beartype


def cplexProblemInterpreter(
    problemDict: dict, optModel, interface, timelimit=1e5, verbose=0
):
    """
    The purpose of this function is to interpret the problem dictionary into cplex, so we can solve it.
    Any problem with the correct format can be solved with this function.

    :param problemDict: Dictionary with the problem information
    :param interface: Interface to use to solve the problem, this is automatically set by the function setSolver.
        See the :func:`gMCSpy.Utilities.setSolver` function for more information.

    :returns:
        - **problemInterpreted:** An cplex model that contains the solver and the problem to solve.
    """

    # Setting cplex parameters
    logName = f"{problemDict.logPath}/solverLog/cplex_{problemDict.getModelName()}_{problemDict.getCharacteristicParameter()}.log"
    if verbose == 2:
        filename = createLog(logName)
        cplexlog = open(filename, "w")
        optModel.set_log_stream(cplexlog)
        optModel.set_error_stream(cplexlog)
        optModel.set_warning_stream(cplexlog)
        optModel.set_results_stream(cplexlog)
    elif verbose == 0:
        optModel.set_log_stream(None)
        optModel.set_error_stream(None)
        optModel.set_warning_stream(None)
        optModel.set_results_stream(None)

    # Transform the problem bounds to cplex readables
    infinity = interface.infinity

    variables = problemDict.getVariables()

    for key, var in variables.items():
        optModel.variables.add(names=[var["name"]])
        if var["lb"] == "infinity":
            var["lb"] = -infinity
        if var["ub"] == "infinity":
            var["ub"] = infinity
        optModel.variables.set_lower_bounds(var["name"], var["lb"])
        optModel.variables.set_upper_bounds(var["name"], var["ub"])

        if var["vtype"] == "binary":
            optModel.variables.set_types(var["name"], optModel.variables.type.binary)

        elif var["vtype"] == "continuous":
            optModel.variables.set_types(
                var["name"], optModel.variables.type.continuous
            )

    if verbose > 0:
        print("Variables added")

    for key, data in problemDict.getConstraints().items():
        variableNamesCon = [variables[j]["name"] for j in data["variables"]]
        coefficients = data["coefficients"]
        sense = data["sense"]
        optModel.linear_constraints.add(
            lin_expr=[interface.SparsePair(ind=variableNamesCon, val=coefficients)],
            rhs=[data["rhs"]],
            senses=[sense],
            names=[key],
        )

    if verbose > 0:
        print("Constraints added")

    if problemDict.getIndicators():
        for key, data in problemDict.getIndicators().items():
            indicatorVariable = variables[data["indicatorVariable"]]["name"]
            variable = [variables[data["variable"]]["name"]]

            if data["activeWhen"] == 1:
                complemented = 0
                senseCon = data["sense"]

            else:
                complemented = 1
                senseCon = data["sense"]

            expr = interface.SparsePair(ind=variable, val=np.ones(len(variable)))
            optModel.indicator_constraints.add(
                indvar=indicatorVariable,
                complemented=complemented,
                lin_expr=expr,
                rhs=data["b"],
                sense=senseCon,
                name=key,
            )
        if verbose > 0:
            print("Indicators added")

    # Define objective function of the model
    if problemDict.getSense() == "maximize":
        optModel.objective.set_sense(optModel.objective.sense.maximize)
    elif problemDict.getSense() == "minimize":
        optModel.objective.set_sense(optModel.objective.sense.minimize)
    else:
        raise ValueError("Objective sense not recognized")

    objective = problemDict.getObjective()

    for i, coeff in enumerate(objective["coefficients"]):
        optModel.objective.set_linear(objective["variables"][i], int(coeff))

    if verbose > 0:
        print("Objective Function added")

    return optModel


def gurobiProblemInterpreter(
    problemDict: dict, env, interface, timelimit=1e5, verbose=0
):
    """
    The purpose of this function is to interpret the problem dictionary into gurobi, so we can solve it.
    Any problem with the correct format can be solved with this function.

    :param problemDict: Dictionary with the problem information
    :param interface: Interface to use to solve the problem, this is automatically set by the function setSolver.
        See the :func:`gMCSpy.Utilities.setSolver` function for more information.

    :returns:
        - **problemInterpreted:** An gurobi model that contains the solver and the problem to solve.
    """


    logName = f"{problemDict.logPath}/solverLog/gurobi_{problemDict.getModelName()}_{problemDict.getCharacteristicParameter()}.log"

    # verbose control
    if verbose == 2:
        env.setParam("OutputFlag", 1)
        env.setParam("LogToConsole", 0)
        filename = createLog(logName)
        env.setParam("LogFile", filename)
    elif verbose == 0:
        env.setParam("OutputFlag", 0)
    else:
        env.setParam("OutputFlag", 1)

    env.start()

    optModel = interface.Model(name="MCS", env=env)

    # Transform the problem bounds to gurobi readables
    infinity = interface.GRB.INFINITY

    # change "infinity" to value of infinity

    for key, var in problemDict.getVariables().items():
        if var["vtype"] == "binary":
            optModel.addVar(
                name=var["name"],
                vtype=interface.GRB.BINARY,
                lb=var["lb"],
                ub=var["ub"],
            )
        elif var["vtype"] == "continuous":
            if var["lb"] == "infinity":
                var["lb"] = -infinity
            if var["ub"] == "infinity":
                var["ub"] = infinity
            optModel.addVar(
                name=var["name"],
                vtype=interface.GRB.CONTINUOUS,
                lb=var["lb"],
                ub=var["ub"],
            )

    optModel.update()
    if verbose > 0:
        print("Variables added")

    for key, data in problemDict.getConstraints().items():
        variableNames = [optModel.getVars()[j] for j in data["variables"]]
        expr = interface.LinExpr(data["coefficients"], variableNames)
        if data["sense"] == "E":
            senseCon = interface.GRB.EQUAL
        elif data["sense"] == "G":
            senseCon = interface.GRB.GREATER_EQUAL
        elif data["sense"] == "L":
            senseCon = interface.GRB.LESS_EQUAL

        optModel.addLConstr(expr, senseCon, data["rhs"], name=key)

    optModel.update()
    if verbose > 0:
        print("Constraints added")

    if problemDict.getIndicators():
        for key, data in problemDict.getIndicators().items():
            indicatorVariable = optModel.getVars()[data["indicatorVariable"]]
            variable = optModel.getVars()[data["variable"]]
            exprCon = interface.LinExpr(variable)

            if data["sense"] == "G":
                senseCon = interface.GRB.GREATER_EQUAL
            else:
                senseCon = interface.GRB.LESS_EQUAL

            optModel.addGenConstrIndicator(
                indicatorVariable, data["activeWhen"], exprCon, senseCon, data["b"], key
            )

        if verbose > 0:
            print("Indicators added")

    optModel.update()

    # Define objective function of the model
    if problemDict.getSense() == "maximize":
        optModel.ModelSense = interface.GRB.MAXIMIZE
    elif problemDict.getSense() == "minimize":
        optModel.ModelSense = interface.GRB.MINIMIZE
    else:
        raise ValueError("Objective sense not recognized")

    objective = problemDict.getObjective()
    objectiveVars = [optModel.getVars()[i] for i in objective["variables"]]
    objtiveExpr = interface.LinExpr(objective["coefficients"], objectiveVars)

    optModel.setObjective(objtiveExpr)
    optModel.update()
    if verbose > 0:
        print("Objective Function added")
   
    return optModel


def scipProblemInterpreter(
    problemDict: dict,
    interface,
    timelimit=1e5,
    verbose=0,
):
    model = interface.Model("MCS")
    modelVars = {}
    for key, var in problemDict.getVariables().items():
        if var["vtype"] == "binary":
            modelVars[key] = model.addVar(
                name=var["name"],
                vtype="B",
                lb=var["lb"],
                ub=var["ub"],
            )
        elif var["vtype"] == "continuous":
            if var["lb"] == "infinity" and var["ub"] == "infinity":
                lb = None
                ub = None
            elif var["ub"] == "infinity":
                ub = None
                lb = var["lb"]
            elif var["lb"] == "infinity":
                lb = None
                ub = var["ub"]
            else:
                ub = var["ub"]
                lb = var["lb"]
            modelVars[key] = model.addVar(
                name=var["name"],
                vtype="C",
                lb=lb,
                ub=ub,
            )

    if verbose > 0:
            print("Variables added")
    
    for key, constraint in problemDict.getConstraints().items():
        if len(constraint['variables']) == 0:
            Warning(f"Constraint {key} has no variables")
            continue
        
        if len(constraint['coefficients']) == 0:
            Warning(f"Constraint {key} has no coefficients")
            continue
        
        if constraint["sense"] == "E":
            sense = "=="
        elif constraint["sense"] == "G":
            sense = ">="
        elif constraint["sense"] == "L":
            sense = "<="
        constraint_string = f"{' + '.join([f'{coeff}*modelVars[{var}]' for var, coeff in zip(constraint['variables'], constraint['coefficients'])])} {sense} {constraint['rhs']}"
        code = f"model.addCons({constraint_string}, modifiable=True, name='{key}')"
        exec(code)
    
    if verbose > 0:                  
        print("Constraints added")
        
    if len(problemDict.getIndicators()) > 0:
        for key, indicator in problemDict.getIndicators().items():
            indicatorVariable = indicator["indicatorVariable"]
            if indicator["sense"] == "G":
                sense = ">="
            elif indicator["sense"] == "L":
                sense = "<="
            elif indicator["sense"] == "E":
                sense = "=="
            if indicator["activeWhen"] == 1:
                active = True
            else:
                active = False
            constraint_string = f"{' + '.join([f'modelVars[{var}]' for var in [indicator['variable']]])} {sense} {indicator['b']}"
            
            code = f"model.addConsIndicator({constraint_string}, binvar=modelVars[{indicatorVariable}], activeone={active}, name='{key}')"
            exec(code)
    
    if verbose > 0:        
        print("Indicators added")
    objective = problemDict.getObjective()
    
    constraint_string = f"{' + '.join([f'{coeff}*modelVars[{var}]' for var, coeff in zip(objective['variables'], objective['coefficients'])])}"
    code = f"model.setObjective({constraint_string},'{problemDict.getSense()}')"
    exec(code)
    
    if verbose > 0:
        print("Objective added")
    
    model.hideOutput()
    return model


import numpy as np
from cobra.util.array import create_stoichiometric_matrix
from scipy.sparse import eye, hstack

from .ProblemDefinitions import cobra
from .ProblemDefinitions import List

import logging
import time
import uuid
from tqdm import tqdm


"""
The `MinimalCutSetProblem` class is a Python class that represents a minimal cut set problem. 
It provides methods for preparing constraint matrices, building indicators into the problem,
iterating over a solver, and getting corresponding variables.

Example Usage:
    # Create an instance of the MinimalCutSetProblem class
    problem = MinimalCutSetProblem()

    # Prepare constraint matrices
    cobraModel = ...
    constraintMatrices = problem.prepareConstraintMatrices(cobraModel)

    # Build indicators into the problem
    prefix = ...
    variables = ...
    indicatorVariables = ...
    sense = ...
    constant = ...
    activeWhen = ...
    problem.buildIndicatorsIntoProblem(prefix, variables, indicatorVariables, sense, constant, activeWhen)

    # Iterate over a solver
    problemInterpreted = ...
    solver = ...
    solutionVariables = ...
    bounds = ...
    solutionDict = problem.iterateOverSolver(problemInterpreted, solver, solutionVariables, bounds)

    # Get corresponding variables
    realVariables = ...
    interpretedSolutions = problem.getCorrespondingVariables(solutionDict, realVariables)

Main functionalities:
- Preparing constraint matrices for a given cobra model
- Building indicators into the problem
- Iterating over a solver to find solutions
- Getting corresponding variables for the solutions

Methods:
- prepareConstraintMatrices(cobraModel, companionType=None, companionMatrix=None): Prepares constraint matrices for a given cobra model.
- buildIndicatorsIntoProblem(problem, prefix, variables, indicatorVariables, sense, constant, activeWhen): Builds indicators into the problem.
- iterateOverSolver(problem, problemInterpreted, solver, solutionVariables, bounds, mergeSolutions=False, earlyStop=False, removeRelated=False, verbose=0, handler=None, iterationProgress=False): Iterates over a solver to find solutions.
- getCorrespondingVariables(problem, solutions, realVariables, mergeSolutions=False): Gets corresponding variables for the solutions.

Fields:
- forceLength: A boolean field indicating whether to force a specific solution length.
- verbose: An integer field indicating the verbosity level.
- solver: A string field indicating the solver to use.
- constantB: A float field representing a constant value.
- constantC: A float field representing a constant value.
- cobraModel: A field representing a cobra model.
"""
class MinimalCutSetProblem:
    def __init__(self, forceLength=False, verbose=0, solver=None, B=1e-3, C=1e-3) -> None:
        self.forceLength = forceLength
        self.verbose = verbose
        self.solver = solver
        self.constantB = B
        self.constantC = C
        self.cobraModel = None

    @staticmethod
    def prepareConstraintMatrices(cobraModel, companionType=None, companionMatrix=None) -> dict:
        if cobraModel is None:
            raise Exception("No cobra model provided")

        if companionMatrix is None and companionType is None:
            raise Exception("No companion matrix provided, no companion type provided, to build K use companionType='K', to use G provide companionMatrix")

        matrixS = create_stoichiometric_matrix(cobraModel, array_type="lil")

        [_, reactions] = matrixS.shape

        lb, ub = zip(*[(r._lower_bound, r._upper_bound) for r in cobraModel.reactions])
        lb = np.array(lb)
        ub = np.array(ub)

        # Splitting index
        splittingIndex = np.where(lb < 0)[0]

        # Creating companion matrix if K is requested
        if companionType == 'K':
            companionMatrix = eye(reactions, format="lil")

        negativeLbColumns = matrixS[:, splittingIndex]
        matrixS = hstack((matrixS, -negativeLbColumns))

        negativeLbColumns = companionMatrix[:, splittingIndex]
        companionMatrix = hstack((companionMatrix, negativeLbColumns))
        companionMatrix = companionMatrix.tolil()

        constraintMs = {}
        constraintMs["S"] = matrixS
        constraintMs["Companion"] = companionMatrix
        return constraintMs

    @staticmethod
    def buildIndicatorsIntoProblem(problem, prefix, variables, indicatorVariables, sense, constant, activeWhen):
        for j, item in enumerate(indicatorVariables):
            name = f"{prefix}{j}"
            problem.addIndicator(
                name= name,
                indicatorVariable= item,
                activeWhen= activeWhen,
                variable= variables[j],
                coefficients= 1,
                sense= sense,
                b= constant
            )

    @staticmethod
    def iterateOverSolver(problem, problemInterpreted, solver, solutionVariables, bounds,
                          mergeSolutions=False, earlyStop=False, removeRelated=False, verbose=0, handler=None,
                          iterationProgress=False):
        TIMELIMIT = problem.getTimeLimitParameter()
        logger = logging.getLogger(handler)
        interface = problemInterpreted[1]
        problemInterpreted = problemInterpreted[0]
        solutionLength = 0
        solutionCountPerLength = 0
        solutionCount = 0
        solutionDict = {}
        rhsForceLength = 0
        MAX_SOLUTION_LENGTH = bounds['MAX_SOLUTION_LENGTH']
        MAX_SOLUTIONS = bounds['MAX_SOLUTIONS']
        forceLength = bounds['forceLength']
        solutionVariableIndices = solutionVariables['indices']
        solutionVariableNames = solutionVariables['names']

        if solver == "gurobi":
            OPTIMALITY = 2
            if forceLength:
                rhsForceLength = problemInterpreted.getConstrByName('forceLength').RHS
            while(solutionLength <= MAX_SOLUTION_LENGTH and solutionCountPerLength <= MAX_SOLUTIONS and rhsForceLength <= MAX_SOLUTION_LENGTH):
                startTime = time.time()
                problemInterpreted.optimize()
                endTime = time.time()
                numberOfSolutions = problemInterpreted.SolCount
                if numberOfSolutions > 0:
                    solutions = {}
                    for j in range(numberOfSolutions):
                        problemInterpreted.Params.SolutionNumber = j
                        solutions[j] = np.array(problemInterpreted.getAttr("Xn", problemInterpreted.getVars()))

                    for j in range(numberOfSolutions):
                        currentSolution =[]
                        solutionIndex = list(i for i in np.where(solutions[j][solutionVariableIndices] > 0.5)[0])
                        currentSolution = [
                            problemInterpreted.getVars()[solutionVariableIndices[i]]
                            for i in solutionIndex
                        ]
                        names = [var.VarName for var in currentSolution]
                        newConExpr = interface.LinExpr(np.ones(len(currentSolution)), currentSolution)
                        problemInterpreted.addLConstr(
                            newConExpr,
                            interface.GRB.LESS_EQUAL,
                            len(currentSolution) - 1,
                            f"{str(uuid.uuid4())}_MultipleSol",
                        )

                        solutionCountPerLength = solutionCountPerLength + 1
                        currentSolution = [solutionVariableNames[i] for i in solutionIndex]
                        solutionInformation = {}

                        solutionInformation["solution"] = currentSolution
                        if mergeSolutions:
                            solutionInformation["solution"] = frozenset().union(*solutionInformation["solution"])
                        if removeRelated:
                            variablesToBound = problem.getContainedVariables(solutionInformation["solution"])
                            afectedVars = 0
                            for name in variablesToBound:
                                problemInterpreted.getVarByName(name).setAttr('UB', 0)
                                afectedVars += 1
                        solLength = len(solutionInformation["solution"])
                        keyName = "order" + str(solLength) + "_" + str(solutionCountPerLength)
                        solutionInformation["order"] = keyName
                        solutionDict[solutionCount] = solutionInformation
                        if  solLength > solutionLength:
                            solutionLength = len(solutionInformation["solution"])
                        solutionCount += 1

                    timeTaken = endTime - startTime
                    if verbose > 0:
                        print(f"Iteration yielded {numberOfSolutions} solutions of length: {solLength}")
                        logger.info(f"timeOrder{solLength}: {timeTaken}")

                    if earlyStop:
                        if (solutionLength == MAX_SOLUTION_LENGTH) & (abs(timeTaken - TIMELIMIT) > 15 ):
                            logger.info("Optimality reached, early stop")
                            break

                else:
                    if forceLength:
                        rhsForceLength = rhsForceLength + 1
                        problemInterpreted.getConstrByName('forceLength').setAttr("RHS", rhsForceLength)
                    else:
                        break
        elif solver == "cplex":
            OPTIMALITY = 0
            if forceLength:
                rhsForceLength = problemInterpreted.linear_constraints.get_rhs('forceLength')
            while(solutionLength <= MAX_SOLUTION_LENGTH and solutionCountPerLength <= MAX_SOLUTIONS and rhsForceLength <= MAX_SOLUTION_LENGTH):
                startTime = time.time()
                problemInterpreted.populate_solution_pool()
                endTime = time.time()
                numberOfSolutions = problemInterpreted.solution.pool.get_num()
                if numberOfSolutions > 0:
                    solutions = {}
                    for j in range(numberOfSolutions):
                        solutions[j] = np.array(problemInterpreted.solution.pool.get_values(j))

                    for j in range(numberOfSolutions):
                        currentSolution =[]
                        solutionIndex = list(i for i in np.where(solutions[j][solutionVariableIndices] > 0.5)[0])
                        currentSolution = [solutionVariableIndices[i] for i in solutionIndex]
                        names = problemInterpreted.variables.get_names(currentSolution)
                        keyName = "order" + str(len(solutionIndex)) + "_" + str(solutionCountPerLength)
                        solutionCountPerLength = solutionCountPerLength + 1
                        currentSolutionNames = [solutionVariableNames[i] for i in solutionIndex]
                        solutionInformation = {}
                        solutionInformation["solution"] = currentSolutionNames
                        constraintLength = len(currentSolution) - 1 
                        if mergeSolutions:
                            solutionInformation["solution"] = frozenset().union(*solutionInformation["solution"])
                            
                        problemInterpreted.linear_constraints.add(
                        lin_expr=[
                            interface.SparsePair(
                                ind=currentSolution, val=np.ones(len(currentSolution))
                            )
                        ],
                        rhs=[constraintLength],
                        senses=["L"],
                        names=[f"{'-'.join(names)}_MultipleSol"])
                        if removeRelated:
                            variablesToBound = problem.getContainedVariables(solutionInformation["solution"])
                            for name in variablesToBound:
                                problemInterpreted.variables.set_upper_bounds(name, 0)
                        keyName = "order" + str(len(solutionInformation["solution"])) + "_" + str(solutionCountPerLength)
                        solutionInformation["order"] = keyName
                        solutionDict[solutionCountPerLength] = solutionInformation
                        if len(solutionInformation["solution"]) > solutionLength:
                            solutionLength = len(solutionInformation["solution"])
                        solutionCount += 1


                    if verbose > 0:
                        print(f"Iteration yielded {numberOfSolutions} solutions of length: {len(solutionInformation['solution'])}")
                        timeTaken = endTime - startTime
                        logger.info(f"timeOrder{len(solutionInformation['solution'])}: {timeTaken}")
                    
                    if earlyStop:
                        if (solutionLength == MAX_SOLUTION_LENGTH) & (abs(timeTaken - TIMELIMIT) > 15 ):
                            logger.info("Optimality reached, early stop")
                            break
                        
                else:
                    if forceLength:
                        rhsForceLength = rhsForceLength + 1
                        problemInterpreted.linear_constraints.set_rhs('forceLength', rhsForceLength)
                    else:
                        break

        elif solver == "scip":
            if forceLength:
                rhsForceLength = problem.getConstraintByName('forceLength')['rhs']

            while(solutionLength <= MAX_SOLUTION_LENGTH and solutionCountPerLength <= MAX_SOLUTIONS and rhsForceLength <= MAX_SOLUTION_LENGTH):
                notSkip = True
                startTime = time.time()
                try:
                    problemInterpreted.optimize()
                except:
                    print("Error in SCIP")
                    notSkip = False
                endTime = time.time()

     
                if problemInterpreted.getStatus() == "optimal" and notSkip:
                    solution = []
                    variables = []
                    for j in problemInterpreted.getVars():
                        if problemInterpreted.getVal(j) > 0.5 and 'zp_' in j.name:
                            pos = int(j.name.split('_')[1])
                            variables.append(pos)
                            solution.append(solutionVariableNames[solutionVariableIndices.index(pos)])
                            coefficients = [1 for i in variables]

                    problem.addConstraint(f"{str(uuid.uuid4())}_MultipleSol", variables, coefficients, "L", len(variables) - 1)
                    [problemInterpreted, interface] = problem.interpretProblem(verbose=0)
                    #problemInterpreted.writeProblem(f"model_scip.lp")
                    keyName = "order_" + str(len(solution)) + "_" + str(solutionCountPerLength)
                    solutionCountPerLength = solutionCountPerLength + 1
                    solutionInformation = {}
                    solutionInformation["order"] = keyName
                    solutionInformation["solution"] = solution
                    if mergeSolutions:
                        solutionInformation["solution"] = frozenset().union(*solutionInformation["solution"])
                    solLength = len(solutionInformation["solution"])
                    solutionDict[solutionCountPerLength] = solutionInformation
                    if  solLength > solutionLength:
                        solutionLength = len(solutionInformation["solution"])

                    if verbose > 0:
                        timeTaken = endTime - startTime
                        print(f"Solution of length {solLength} found in {timeTaken} seconds")
                        logger.info(f"timeOrder{solLength}: {timeTaken}")

                else:
                    if forceLength:
                        rhsForceLength = rhsForceLength + 1
                        problem.getConstraintByName('forceLength')['rhs'] = rhsForceLength
                        [problemInterpreted, interface] = problem.interpretProblem(verbose=0)
                    else:
                        break
        return solutionDict

    @staticmethod
    def getCorrespondingVariables(problem, solutions, realVariables, mergeSolutions=False):
        variables = list(problem.getVariableNames().keys())

        zpVariables = [variables[var] for var in problem.getObjective()['variables']]


        #The variables that correspond to the problem variables, e.g the real variables, reactions, genes, Knockouts, etc. 
        interpretedSolutions = {}
        #Match the position of the zp_i variables in the solution
        for j, key in enumerate(solutions):
            solution= solutions[key]
            indices = [zpVariables.index(variable) for variable in solution]
            #Get the corresponding real variables
            solutionInformation = {}
            solutionInformation["order"] = key
            solutionInformation["solution"] = [realVariables[index] for index in indices]
            if mergeSolutions:
                solutionInformation["solution"] = frozenset().union(*solutionInformation["solution"])
            solutionInformation["length"] = len(solution)
            interpretedSolutions[j] = solutionInformation


        return interpretedSolutions
    
    

def prepareModelNutrientGeneMCS(cobraModel: cobra.Model, exchangeReactions: List):
            
    # Perform a reaction splitting, we are only interested in the exchange reactions that bring metabolites into the system (inputs)
    
    # Check what reactions are reversible
    if exchangeReactions is None:
        listOfGeneRules = []
        for reaction in cobraModel.exchanges:
            if len(reaction.metabolites) > 1:
                print("WARNING: Some exchange reactions exchange more than one metabolite.")
            name  = list(reaction.metabolites.keys())[0].id + "_ag"
            if reaction.reversibility:
                # Split the reaction into two reactions, one for the forward direction and one for the reverse direction
                newReaction = cobra.Reaction(reaction.id + "_reverse")
                newReaction.name = reaction.id + "_reverse"
                newReaction.lower_bound = 0
                newReaction.upper_bound = 1000
                for metabolite, value in reaction.metabolites.items():
                    newReaction.add_metabolites({metabolite: -value})     
                
                # Check if reaction is input, if from nothing yields a product, then it is an input
                if len(newReaction.reactants) == 0 and len(newReaction.products) == 1:
                    newReaction.gene_reaction_rule = name
                    listOfGeneRules.append(newReaction.id)
                
                # Last step add reaction to model 
                cobraModel.add_reactions([newReaction])
                
                reaction.lower_bound = 0
                if len(reaction.reactants) == 0 and len(reaction.products) == 1:
                    reaction.gene_reaction_rule = name
                    listOfGeneRules.append(reaction.id)
            else:
                if len(reaction.reactants) == 0 and len(reaction.products) == 0:
                    reaction.gene_reaction_rule = name
                    listOfGeneRules.append(reaction.id)
                    



    print('===================================================================')
    
    
    return cobraModel



from .ProblemInterpreters import np
from .Utilities import beartype
import scipy
import cobra
from beartype.typing import List
import math

from .MinimalCutSetClass import MinimalCutSetProblem
from .OptimizationProblem import OptimizationProblem


def buildDictionaryMCSProblem(
    cobraModel: cobra.Model,
    forceLength: bool,
    rxnToForce: List[str],
    verbose: int = 0,
    targetB: float = 1e-3,
    cConstant: float = 1e-3,
):
    """
    Builds the dictionary that defines the MILP problem to be solved by the MCS algorithm.

    :param modelStruct: Cobra Model, the metabolic model to be operated on. Only Cobra models are supported.
    :param forceLength: boolean. If true, the length of the solution is forced to be off a given length each iteration, otherwise the length is not forced.
    :param targetB: float. Desired activity level of the metabolic task to be disrupted.
    :param cConstant: float.  Used to activate w variable
    :param rxnToForce: list of strings. Reactions to be forced to be part of the solution. Empty by default.

    If the bounds are left as None they are ment to be infinity (None = Infinity), so each solver can interpret it as it is defined.

    :return: problemDict. Dictionary that defines the MILP problem to be solved by any solver defined in gMCSpy.
    """

    problemMethods = MinimalCutSetProblem()
    problem = OptimizationProblem()

    # Alpha. Used to relate the lower bound of v variables with z variables
    ALPHA = 1

    matrices = problemMethods.prepareConstraintMatrices(cobraModel, "K")

    matrixS = matrices["S"]
    matrixK = matrices["Companion"]

    # S matrix changed size
    [metabolites, reactions] = matrixS.shape

    # Objective vector of model
    objectiveVector = [r.objective_coefficient for r in cobraModel.reactions]
    indexC = objectiveVector.index(1)

    # Target vector
    t = np.zeros(reactions)
    t[indexC] = 1

    if rxnToForce:
        if verbose > 0:
            print("Forcing reactions to be part of the solution: " + str(rxnToForce))
        # Find the index of the reactions in the model
        rxnIndex = [
            cobraModel.reactions.index(cobraModel.reactions.get_by_id(rxn))
            for rxn in rxnToForce
        ]
        matrixK = matrixK[rxnIndex, :]
        reactionsK = matrixK.shape[0]

    # Add variables to the problem

    vIndex = {}
    vIndex["u"] = list(range(0, metabolites))
    vIndex["vp"] = list(range(vIndex["u"][-1] + 1, vIndex["u"][-1] + reactionsK + 1))
    vIndex["w"] = [vIndex["vp"][-1] + 1]
    vIndex["zp"] = list(range(vIndex["w"][-1] + 1, vIndex["w"][-1] + reactionsK + 1))
    vIndex["zw"] = [vIndex["zp"][-1] + 1]
    name = "u_"
    for u in vIndex["u"]:
        problem.addVariable(
            index=u, prefix=name, lb="infinity", ub="infinity", vtype="continuous"
        )

    name = "vp_"
    for vp in vIndex["vp"]:
        problem.addVariable(
            index=vp, prefix=name, lb=0, ub="infinity", vtype="continuous"
        )

    name = "w_"
    w = vIndex["w"][0]
    problem.addVariable(index=w, prefix=name, lb=0, ub="infinity", vtype="continuous")

    name = "zp_"
    count = 0
    for zp in vIndex["zp"]:
        problem.addVariable(
            index=zp, prefix=name, lb=0, ub=1, vtype="binary", relatedTo=count
        )
        count += 1

    name = "zw_"
    zw = vIndex["zw"][0]
    problem.addVariable(index=zw, prefix=name, lb=0, ub=1, vtype="binary")

    vGroups = {}
    vGroups["v"] = vIndex["vp"] + vIndex["w"]
    vGroups["z"] = vIndex["zp"] + vIndex["zw"]

    cIndex = {}

    cIndex["Ndual"] = list(range(0, reactions))
    cIndex["forceBioCons"] = [cIndex["Ndual"][-1] + 1]
    cIndex["forceLength"] = [cIndex["forceBioCons"][-1] + 1]

    matrixS = matrixS.transpose().tocsr()
    matrixK = matrixK.transpose().tocsr()
    tIndexConstraint = np.where(t != 0)[0]
    for i in cIndex["Ndual"]:
        constraintName = f"Ndual_constraint{i}"
        variableIndicesU = matrixS[i].indices
        vars = []
        vars = [vIndex["u"][j] for j in variableIndicesU]
        variableIndicesVp = matrixK[i].indices
        vars.extend([vIndex["vp"][j] for j in variableIndicesVp])
        coefs = np.concatenate((matrixS[i].data, matrixK[i].data))
        problem.addConstraint(
            name=constraintName, variables=vars, coefficients=coefs, sense="G", rhs=0
        )

    constraintName = f"Ndual_constraint{tIndexConstraint[0]}"
    constraint = problem.getConstraint(constraintName)
    constraint["variables"].extend([w])
    coefs = np.concatenate((constraint["coefficients"], -t[tIndexConstraint]))
    problem.modifyConstraint(
        constraintName=constraintName,
        variables=constraint["variables"],
        coefficients=coefs,
        sense=constraint["sense"],
        rhs=constraint["rhs"],
    )

    problem.addConstraint(
        name="forceBioCons",
        variables=[w],
        coefficients=[-targetB],
        sense="E",
        rhs=-cConstant,
    )

    if forceLength:
        variables = vIndex["zp"]
        problem.addConstraint(
            name="forceLength",
            variables=variables,
            coefficients=np.ones(len(variables)),
            sense="E",
            rhs=1,
        )

    problem.addObjective(vIndex["zp"], np.ones(len(vIndex["zp"])), sense="minimize")

    # Indicators
    # z = 1  -->  v >= alpha
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOn_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        activeWhen=1,
        sense="G",
        constant=ALPHA,
    )

    # z = 0  -->  v <= 0
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOff_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        activeWhen=0,
        sense="L",
        constant=0,
    )

    return problem


def buildDictionaryMCSTargetedProblem(
    cobraModel: cobra.Model,
    forceLength: bool,
    rxnSubset: List[str],
    rxnKnockOut: List[str],
    verbose: int = 0,
    targetB: float = 1e-3,
    cConstant: float = 1e-3,
):
    """
    Builds the dictionary that defines the MILP problem to be solved by the MCS algorithm.

    :param modelStruct: Cobra Model, the metabolic model to be operated on. Only Cobra models are supported.
    :param forceLength: boolean. If true, the length of the solution is forced to be off a given length each iteration, otherwise the length is not forced.
    :param targetB: float. Desired activity level of the metabolic task to be disrupted.
    :param cConstant: float.  Used to activate w variable
    :param rxnToForce: list of strings. Reactions to be forced to be part of the solution. Empty by default.

    If the bounds are left as None they are ment to be infinity (None = Infinity), so each solver can interpret it as it is defined in each solver.

    :return: problemDict. Dictionary that defines the MILP problem to be solved by any solver defined in gMCSpy.
    """

    problemMethods = MinimalCutSetProblem()
    problem = OptimizationProblem()

    # Alpha. Used to relate the lower bound of v variables with z variables
    ALPHA = 1e-3

    matrices = problemMethods.prepareConstraintMatrices(cobraModel, "K")

    matrixS = matrices["S"]
    matrixK = matrices["Companion"]

    # S matrix changed size
    [mUpdated, nUpdated] = matrixS.shape

    # Objective vector of model
    objectiveVector = [r.objective_coefficient for r in cobraModel.reactions]
    indexC = objectiveVector.index(1)

    # Target vector
    t = np.zeros(nUpdated)
    t[indexC] = 1

    if rxnSubset:
        if verbose > 0:
            print("Forcing reactions to be part of the solution: " + str(rxnSubset))
        # Find the index of the reactions in the model
        rxnIndex = [
            cobraModel.reactions.index(cobraModel.reactions.get_by_id(rxn))
            for rxn in rxnSubset
        ]
        # Order the reactions to be forced by the index of the reactions in the model
        rxnIndex = sorted(rxnIndex)
        matrixK = matrixK[rxnIndex, :]
        n = matrixK.shape[0]

    rxnKnockOutIndex = np.zeros(len(cobraModel.reactions))
    rxnKnockOutIndex[
        [
            cobraModel.reactions.index(cobraModel.reactions.get_by_id(rxn))
            for rxn in rxnKnockOut
        ]
    ] = 1

    vIndex = {}
    vGroups = {}
    vIndex["u"] = list(range(0, mUpdated))
    vIndex["vp"] = list(range(vIndex["u"][-1] + 1, vIndex["u"][-1] + n + 1))
    vIndex["w"] = [vIndex["vp"][-1] + 1]
    vIndex["zp"] = list(range(vIndex["w"][-1] + 1, vIndex["w"][-1] + n + 1))
    vIndex["zw"] = [vIndex["zp"][-1] + 1]
    vIndex["epsp"] = list(range(vIndex["zw"][-1] + 1, vIndex["zw"][-1] + n + 1))
    vIndex["espw"] = [vIndex["epsp"][-1] + 1]
    vIndex["delp"] = list(range(vIndex["espw"][-1] + 1, vIndex["espw"][-1] + n + 1))
    vIndex["delw"] = [vIndex["delp"][-1] + 1]
    vIndex["x"] = list(range(vIndex["delw"][-1] + 1, vIndex["delw"][-1] + nUpdated + 2))
    nVars = vIndex["x"][-1] + 1

    for u in vIndex["u"]:
        problem.addVariable(
            index=u, prefix="u_", lb="infinity", ub="infinity", vtype="continuous"
        )

    for vp in vIndex["vp"]:
        problem.addVariable(
            index=vp, prefix="vp_", lb=0, ub="infinity", vtype="continuous"
        )

    for w in vIndex["w"]:
        problem.addVariable(
            index=w, prefix="w_", lb=0, ub="infinity", vtype="continuous"
        )

    for zp in vIndex["zp"]:
        problem.addVariable(index=zp, prefix="zp_", lb=0, ub=1, vtype="binary")

    for zw in vIndex["zw"]:
        problem.addVariable(index=zw, prefix="zw_", lb=0, ub=1, vtype="binary")

    for epsp in vIndex["epsp"]:
        problem.addVariable(
            index=epsp, prefix="epsp_", lb=0, ub="infinity", vtype="continuous"
        )

    for espw in vIndex["espw"]:
        problem.addVariable(index=espw, prefix="espw_", lb=0, ub=0, vtype="continuous")

    for delp in vIndex["delp"]:
        problem.addVariable(
            index=delp, prefix="delp_", lb=0, ub="infinity", vtype="continuous"
        )

    for delw in vIndex["delw"]:
        problem.addVariable(index=delw, prefix="delw_", lb=0, ub=0, vtype="continuous")

    for x in vIndex["x"][:-1]:
        problem.addVariable(
            index=x, prefix="x_", lb=0, ub="infinity", vtype="continuous"
        )
    problem.addVariable(
        index=vIndex["x"][-1], prefix="x_", lb=1000, ub="infinity", vtype="continuous"
    )

    vGroups["v"] = vIndex["vp"] + vIndex["w"]
    vGroups["z"] = vIndex["zp"] + vIndex["zw"]
    vGroups["eps"] = vIndex["epsp"] + vIndex["espw"]
    vGroups["del"] = vIndex["delp"] + vIndex["delw"]

    cIndex = {}

    cIndex["Ndual"] = list(range(0, nUpdated))
    cIndex["forceBioCons"] = [cIndex["Ndual"][-1] + 1]
    cIndex["forceKnockOut"] = [cIndex["forceBioCons"][-1] + 1]

    numberOfLinearCombinations = matrixS.shape[0] + matrixK.shape[0] + 1 + 1
    cIndex["linearCombination"] = list(
        range(
            cIndex["forceKnockOut"][-1] + 1,
            cIndex["forceKnockOut"][-1] + numberOfLinearCombinations,
        )
    )

    cIndex["forceLength"] = [cIndex["linearCombination"][-1] + 1]
    nCons = cIndex["forceLength"][-1] + 1

    A = scipy.sparse.lil_matrix((nCons, nVars))
    A[np.ix_(cIndex["Ndual"], vIndex["u"])] = matrixS.transpose()
    A[np.ix_(cIndex["Ndual"], vIndex["vp"])] = matrixK.transpose()
    A[np.ix_(cIndex["Ndual"], vIndex["w"])] = -t
    A[np.ix_(cIndex["forceBioCons"], vIndex["w"])] = -targetB
    A[np.ix_(cIndex["forceKnockOut"], vIndex["vp"])] = rxnKnockOutIndex

    auxS = scipy.sparse.hstack(
        [matrixS, scipy.sparse.lil_matrix(np.zeros((mUpdated, 1)))]
    )
    auxK = scipy.sparse.hstack([matrixK, scipy.sparse.lil_matrix(np.zeros((n, 1)))])
    auxT = scipy.sparse.hstack([scipy.sparse.lil_matrix(-t), targetB])
    A[np.ix_(cIndex["linearCombination"], vIndex["x"])] = scipy.sparse.vstack(
        [auxS, auxK, auxT]
    )

    matrixEp1 = scipy.sparse.lil_matrix(
        (mUpdated, (len(vIndex["vp"]) + len(vIndex["w"])))
    )
    matrixEp2 = scipy.sparse.eye((len(vIndex["vp"]) + len(vIndex["w"])), format="lil")
    auxEp = scipy.sparse.vstack([matrixEp1, -matrixEp2])
    A[np.ix_(cIndex["linearCombination"], vGroups["eps"])] = auxEp

    A[np.ix_(cIndex["linearCombination"], vGroups["del"])] = -auxEp

    rhs = np.zeros(nCons)
    rhs[cIndex["Ndual"]] = None
    rhs[cIndex["forceBioCons"][0]] = -cConstant
    rhs[cIndex["forceKnockOut"]] = None
    rhs[
        cIndex["linearCombination"][mUpdated : len(cIndex["linearCombination"]) - 1]
    ] = rxnKnockOutIndex

    lhs = np.zeros(nCons)
    lhs[cIndex["Ndual"]] = 0
    lhs[cIndex["forceBioCons"][0]] = -1
    lhs[cIndex["forceKnockOut"]] = 1e-3
    lhs[
        cIndex["linearCombination"][mUpdated : len(cIndex["linearCombination"]) - 1]
    ] = rxnKnockOutIndex

    sense = np.array(["E"] * nCons)
    sense[cIndex["Ndual"]] = "G"
    sense[cIndex["forceBioCons"][0]] = "G"
    sense[cIndex["forceKnockOut"]] = "G"

    A = A.tocsr()
    problem.addConstraintsFromSparseMatrix(A, lhs, sense)

    varsForceBioCons = A[cIndex["forceBioCons"][0]].indices
    coefficients = A[cIndex["forceBioCons"][0]].data
    constraintName = f'c{cIndex["forceBioCons"][0]}'
    problem.modifyConstraint(
        constraintName=constraintName,
        variables=varsForceBioCons,
        coefficients=coefficients,
        sense="L",
        rhs=-cConstant,
    )
    # problem.addConstraint(name="forceBioCons_2",variables=varsForceBioCons, coefficients=coefficients, sense="L", rhs=-cConstant)

    problem.addObjective(vIndex["zp"], np.ones(len(vIndex["zp"])), sense="minimize")

    vGroups["v"] = vIndex["vp"] + vIndex["w"]
    vGroups["z"] = vIndex["zp"] + vIndex["zw"]
    vGroups["eps"] = vIndex["epsp"] + vIndex["espw"]
    vGroups["del"] = vIndex["delp"] + vIndex["delw"]

    # Indicators
    # z = 1  -->  v >= alpha
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOn_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        sense="G",
        constant=ALPHA,
        activeWhen=1,
    )

    # z = 0  -->  v <= 0
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOff_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        sense="L",
        constant=0,
        activeWhen=0,
    )

    # z = 1  -->  epsilon <= 0
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOnEpsilon_",
        variables=vGroups["eps"],
        indicatorVariables=vGroups["z"],
        sense="L",
        constant=0,
        activeWhen=1,
    )

    # z = 0  -->  epsilon <= M
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOffEpsilon_",
        variables=vGroups["eps"],
        indicatorVariables=vGroups["z"],
        sense="L",
        constant=1e3,
        activeWhen=0,
    )

    return problem


@beartype
def buildDictionaryFBA(cobraModel: cobra.Model):
    """Build FBA problem dictionary
    The basic idea of FBA is to use constraints based on the mass balance and stoichiometry of reactions to determine the
    steady-state fluxes through the metabolic network. The steady-state assumption assumes that the concentrations of metabolites
    do not change over time, which is reasonable for many metabolic systems.
    """

    # build Stoichiometric matrix
    matrixS = cobra.util.array.create_stoichiometric_matrix(
        cobraModel, array_type="lil"
    )

    [numberMetabolites, numberReactions] = matrixS.shape
    ## m = number of metabolites
    ## n = number of reactions

    # reactions bounds
    lb, ub = zip(*[(r._lower_bound, r._upper_bound) for r in cobraModel.reactions])
    lb = np.array(lb)
    ub = np.array(ub)

    problem = OptimizationProblem()

    variableNames = {}
    vIndices = {}

    vIndices["u"] = list(range(0, numberReactions))
    nVars = vIndices["u"][-1] + 1

    prefix = "u_"
    for u in vIndices["u"]:
        problem.addVariable(
            index=u, prefix=prefix, lb=lb[u], ub=ub[u], vtype="continuous"
        )

    cIndices = {}

    cIndices["S"] = list(range(0, numberMetabolites))
    nCons = cIndices["S"][-1] + 1
    matrixS = matrixS.tocsr()

    for i in cIndices["S"]:
        constraintName = f"Srow_{i}"
        variablesIndices = matrixS[i].indices
        coefs = matrixS[i].data
        problem.addConstraint(
            name=constraintName,
            variables=variablesIndices,
            coefficients=coefs,
            sense="E",
            rhs=0,
        )

    # Objective vector of model
    objectiveVector = [r.objective_coefficient for r in cobraModel.reactions]
    indexC = [objectiveVector.index(1)]

    problem.addObjective(indexC, np.ones(len(indexC)), sense="maximize")

    return problem


def buildDictionaryFBAWithDeletions(
    cobraModel: cobra.Model, gMCS: List[str] = None, isoformSeparator: str = None
):
    """Build FBA problem dictionary
    The basic idea of FBA is to use constraints based on the mass balance and stoichiometry of reactions to determine the
    steady-state fluxes through the metabolic network. The steady-state assumption assumes that the concentrations of metabolites
    do not change over time, which is reasonable for many metabolic systems.
    """

    # build Stoichiometric matrix
    matrixS = cobra.util.array.create_stoichiometric_matrix(
        cobraModel, array_type="lil"
    )

    [numberMetabolites, numberReactions] = matrixS.shape
    ## m = number of metabolites
    ## n = number of reactions

    if isoformSeparator is not None:
        # Get the GPR of each reaction
        GPRs = [
            [r, r.gpr]
            for r in cobraModel.reactions
            for gene in r.gpr.genes
            if gene.split(isoformSeparator)[0] in gMCS
        ]
        # modelReactions = cobraModel.reactions
        # modelGenes = [g.id for g in cobraModel.genes]
        reactionsToKnockOut = []
        for reaction, gpr in GPRs:
            if len(gpr.genes) == 1:
                reactionsToKnockOut.append(reaction.id)
            else:
                expr = gpr.to_string()
                for gene in gpr.genes:
                    # Substitute the gene name for True in the expression
                    geneSplit = gene.split(isoformSeparator)[0]
                    if geneSplit in gMCS:
                        expr = expr.replace(gene, "False")
                    else:
                        expr = expr.replace(gene, "True")
                # Evaluate the expression
                outcome = eval(expr)
                if not outcome:
                    reactionsToKnockOut.append(reaction.id)

    else:
        # Get the GPR of each reaction
        GPRs = [
            [r, r.gpr]
            for r in cobraModel.reactions
            for gene in gMCS
            if gene in r.gpr.genes
        ]
        # modelReactions = cobraModel.reactions
        # modelGenes = [g.id for g in cobraModel.genes]
        reactionsToKnockOut = []
        for reaction, gpr in GPRs:
            if len(gpr.genes) == 1:
                reactionsToKnockOut.append(reaction.id)
            else:
                expr = gpr.to_string()
                for gene in gpr.genes:
                    # Substitute the gene name for True in the expression
                    if gene in gMCS:
                        expr = expr.replace(gene, "False")
                    else:
                        expr = expr.replace(gene, "True")
                # Evaluate the expression
                outcome = eval(expr)
                if not outcome:
                    reactionsToKnockOut.append(reaction.id)

    """     if isoformSeparator is not None:
        for gene in cobraModel.genes:
            gene.id = gene.id.split(isoformSeparator)[0]
            if gene.id in gMCS:
                for reaction in gene.reactions:
                    reactionsToKnockOut.append(reaction.id)
    else:
        for gene in cobraModel.genes:
            if gene.id in gMCS:
                for reaction in gene.reactions:
                    reactionsToKnockOut.append(reaction.id) """

    reactionsToKnockOut = list(set(reactionsToKnockOut))

    # reactions bounds
    lb = []
    ub = []
    for reaction in cobraModel.reactions:
        if reaction.id in reactionsToKnockOut:
            lb.append(0)
            ub.append(0)
        else:
            lb.append(reaction.lower_bound)
            ub.append(reaction.upper_bound)

    lb = np.array(lb)
    ub = np.array(ub)

    problem = OptimizationProblem()

    variableNames = {}
    vIndices = {}

    vIndices["u"] = list(range(0, numberReactions))
    nVars = vIndices["u"][-1] + 1

    prefix = "u_"
    for u in vIndices["u"]:
        problem.addVariable(
            index=u, prefix=prefix, lb=lb[u], ub=ub[u], vtype="continuous"
        )

    cIndices = {}

    cIndices["S"] = list(range(0, numberMetabolites))
    nCons = cIndices["S"][-1] + 1
    matrixS = matrixS.tocsr()

    for i in cIndices["S"]:
        constraintName = f"Srow_{i}"
        variablesIndices = matrixS[i].indices
        coefs = matrixS[i].data
        problem.addConstraint(
            name=constraintName,
            variables=variablesIndices,
            coefficients=coefs,
            sense="E",
            rhs=0,
        )

    # Objective vector of model
    objectiveVector = [r.objective_coefficient for r in cobraModel.reactions]
    indexC = [objectiveVector.index(1)]

    problem.addObjective(indexC, np.ones(len(indexC)), sense="maximize")

    return problem


def buildDictionaryMCSGeneProblem(
    cobraModel: cobra.Model,
    gDict: dict,
    matrixG: scipy.sparse.csc_matrix,
    relationships: dict,
    maxKOLength: int,
    numberNewGenesByKO: dict,
    minKOLength: int = 0,
    forceLength: bool = False,
    verbose: int = 0,
    targetB: float = 1e-3,
    cConstant: float = 1e-3,
    logPath: str = None,
):
    """
    Builds the dictionary that defines the MILP problem to be solved by the MCS algorithm.

    :param modelStruct: Cobra Model, the metabolic model to be operated on. Only Cobra models are supported.
    :param forceLength: boolean. If true, the length of the solution is forced to be off a given length each iteration, otherwise the length is not forced.
    :param targetB: float. Desired activity level of the metabolic task to be disrupted.
    :param cConstant: float.  Used to activate w variable
    :param rxnToForce: list of strings. Reactions to be forced to be part of the solution. Empty by default.

    If the bounds are left as None they are ment to be infinity (None = Infinity), so each solver can interpret it as it is defined.

    :return: problemDict. Dictionary that defines the MILP problem to be solved by any solver defined in gMCSpy.
    """

    problem = OptimizationProblem()
    problemMethods = MinimalCutSetProblem()

    # Alpha. Used to relate the lower bound of v variables with z variables
    ALPHA = 1

    numberOfPossibleKO = matrixG.shape[0]
    possibleKOs = list(gDict.keys())

    if relationships is not None:
        numberOfRelationships = sum(len(lst) for lst in relationships.values())
    else:
        numberOfRelationships = 0

    matrices = problemMethods.prepareConstraintMatrices(
        cobraModel=cobraModel, companionMatrix=matrixG
    )
    matrixS = matrices["S"]
    matrixG = matrices["Companion"]

    # S matrix changed size
    [metabolites, reactions] = matrixS.shape

    # Objective vector of model
    objectiveVector = [r.objective_coefficient for r in cobraModel.reactions]
    indexC = objectiveVector.index(1)

    # Target vector
    t = np.zeros(reactions)
    t[indexC] = 1

    # Build variables
    vGroups = {}
    vIndex = {}

    vIndex["u"] = list(range(0, metabolites))
    vIndex["vp"] = list(
        range(vIndex["u"][-1] + 1, vIndex["u"][-1] + numberOfPossibleKO + 1)
    )
    vIndex["w"] = [vIndex["vp"][-1] + 1]
    vIndex["zp"] = list(
        range(vIndex["w"][-1] + 1, vIndex["w"][-1] + numberOfPossibleKO + 1)
    )
    vIndex["zw"] = [vIndex["zp"][-1] + 1]

    name = "u_"
    for u in range(0, metabolites):
        problem.addVariable(
            index=u, prefix=name, lb="infinity", ub="infinity", vtype="continuous"
        )

    name = "vp_"
    for i, vp in enumerate(range(u + 1, u + numberOfPossibleKO + 1)):
        problem.addVariable(
            index=vp,
            prefix=name,
            lb=0,
            ub="infinity",
            vtype="continuous",
            relatedTo=possibleKOs[i],
        )

    name = "w_"
    w = vp + 1
    problem.addVariable(index=w, prefix=name, lb=0, ub="infinity", vtype="continuous")

    name = "zp_"
    for i, zp in enumerate(range(w + 1, w + numberOfPossibleKO + 1)):
        problem.addVariable(
            index=zp, prefix=name, lb=0, ub=1, vtype="binary", relatedTo=possibleKOs[i]
        )

    name = "zw_"
    zw = zp + 1
    problem.addVariable(index=zw, prefix=name, lb=0, ub=1, vtype="binary")

    vGroups["v"] = vIndex["vp"] + vIndex["w"]
    vGroups["z"] = vIndex["zp"] + vIndex["zw"]

    # Build constraints
    cIndex = {}
    cIndex["Ndual"] = list(range(0, reactions))
    cIndex["forceBioCons"] = [cIndex["Ndual"][-1] + 1]
    if numberOfRelationships > 0:
        cIndex["relationships"] = list(
            range(
                cIndex["forceBioCons"][-1] + 1,
                cIndex["forceBioCons"][-1] + numberOfRelationships + 1,
            )
        )
        cIndex["forceLength"] = [cIndex["relationships"][-1] + 1]
    else:
        cIndex["forceLength"] = [cIndex["forceBioCons"][-1] + 1]

    matrixS = matrixS.transpose().tocsr()
    matrixG = matrixG.transpose().tocsr()
    tIndexConstraint = np.where(t != 0)[0]
    for i in cIndex["Ndual"]:
        constraintName = f"Ndual_constraint{i}"
        variableIndicesU = matrixS[i].indices
        vars = [vIndex["u"][j] for j in variableIndicesU]
        variableIndicesVp = matrixG[i].indices
        vars.extend([vIndex["vp"][j] for j in variableIndicesVp])
        coefs_s = matrixS[i].data
        coefs_g = matrixG[i].data
        coefs = np.concatenate((coefs_s, coefs_g))

        problem.addConstraint(constraintName, vars, coefs, "G", 0)

    constraintName = f"Ndual_constraint{tIndexConstraint[0]}"
    constraint = problem.getConstraint(constraintName)
    constraint["variables"].extend([w])
    coefs = np.concatenate((constraint["coefficients"], -t[tIndexConstraint]))
    problem.modifyConstraint(
        constraintName=constraintName,
        variables=constraint["variables"],
        coefficients=coefs,
        sense=constraint["sense"],
        rhs=constraint["rhs"],
    )

    vpVariableNames = {}
    for index, name in zip(vIndex["zp"], gDict.keys()):
        vpVariableNames[name] = index

    if numberOfRelationships > 0:
        subIndex = 0
        for key, values in relationships.items():
            vpContainer = vpVariableNames[key]
            for value in values:
                constraintName = f"relationship_constraint{subIndex}"
                vpContained = vpVariableNames[value]
                vars = [vpContainer, vpContained]
                coefs = [-1, 1]
                problem.addConstraint(constraintName, vars, coefs, "G", 0)
                subIndex += 1

    # Build objective
    objectiveVars = [
        vIndex["zp"][i]
        for i in range(0, len(possibleKOs))
    ]
    problem.addObjective(objectiveVars, list(numberNewGenesByKO.values()), sense="minimize")

    if forceLength:
        constraintName = "forceLength"
        vars = objectiveVars
        coefs = list(numberNewGenesByKO.values())
        problem.addConstraint(constraintName, vars, coefs, "L", 1)

    constraintName = "forceBioCons_constraint"
    vars = vIndex["w"]
    coefs = [-targetB]
    problem.addConstraint(constraintName, vars, coefs, "L", -cConstant)

    # Indicators
    # z = 1  -->  v >= alpha
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOn_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        activeWhen=1,
        sense="G",
        constant=ALPHA,
    )

    # z = 0  -->  v <= 0
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOff_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        activeWhen=0,
        sense="L",
        constant=0,
    )

    return problem


def buildDictionaryMCSGeneTargetedProblem(
    cobraModel: cobra.Model,
    gDict: dict,
    matrixG: scipy.sparse.csc_matrix,
    relationships: dict,
    maxKOLength: int,
    numberNewGenesByKO: dict,
    geneSubset: list,
    targetKOs: list,
    minKOLength: int = 0,
    forceLength: bool = False,
    verbose: int = 0,
    targetB: float = 1e-3,
    cConstant: float = 100,
):
    """
    Builds the dictionary that defines the MILP problem to be solved by the MCS algorithm.

    :param modelStruct: Cobra Model, the metabolic model to be operated on. Only Cobra models are supported.
    :param forceLength: boolean. If true, the length of the solution is forced to be off a given length each iteration, otherwise the length is not forced.
    :param targetB: float. Desired activity level of the metabolic task to be disrupted.
    :param cConstant: float.  Used to activate w variable
    :param rxnToForce: list of strings. Reactions to be forced to be part of the solution. Empty by default.

    If the bounds are left as None they are meant to be infinity (None = Infinity), so each solver can interpret it as it is defined.

    :return: problemDict. Dictionary that defines the MILP problem to be solved by any solver defined in gMCSpy.
    """

    problem = OptimizationProblem()
    problemMethods = MinimalCutSetProblem()

    # Alpha. Used to relate the lower bound of v variables with z variables
    ALPHA = 1

    numberOfPossibleKO = matrixG.shape[0]
    possibleKOs = list(gDict.keys())

    if relationships is not None:
        numberOfRelationships = sum(len(lst) for lst in relationships.values())
    else:
        numberOfRelationships = 0

    matrices = problemMethods.prepareConstraintMatrices(
        cobraModel=cobraModel, companionMatrix=matrixG
    )
    matrixS = matrices["S"]
    matrixG = matrices["Companion"]

    # Find maximum value of matrixS
    maxS = np.max(matrixS.data)
    maxS = 10 ** math.ceil(math.log10(maxS))
    # targetB should 5 orders of manitude smaller than the maximum value of matrixS
    targetB = maxS * 1e-8

    # S matrix changed size
    [metabolites, reactions] = matrixS.shape

    # Objective vector of model
    objectiveVector = [r.objective_coefficient for r in cobraModel.reactions]
    indexC = objectiveVector.index(1)

    # Target vector
    t = np.zeros(reactions)
    t[indexC] = 1

    if geneSubset:
        if verbose > 0:
            print("The solution space is made up by these genes: " + str(geneSubset))

        # Find the interventions that contain the genes in the geneSubset

        interventionsToKeep = []
        for gene in geneSubset:
            for intervention in possibleKOs:
                if gene.issubset(intervention):
                    interventionsToKeep.append(intervention)

        interventionsIndex = [
            possibleKOs.index(intervention) for intervention in interventionsToKeep
        ]
        interventionsIndex = sorted(list(set(interventionsIndex)))
        matrixG = matrixG[interventionsIndex, :]
        n = matrixG.shape[0]

    targetKOIndex = np.zeros(len(possibleKOs))
    for gene in targetKOs:
        for intervention in possibleKOs:
            if gene == intervention:
                targetKOIndex[possibleKOs.index(intervention)] = 1

    vIndex = {}
    vGroups = {}
    vIndex["u"] = list(range(0, metabolites))
    vIndex["vp"] = list(range(vIndex["u"][-1] + 1, vIndex["u"][-1] + n + 1))
    vIndex["w"] = [vIndex["vp"][-1] + 1]
    vIndex["zp"] = list(range(vIndex["w"][-1] + 1, vIndex["w"][-1] + n + 1))
    vIndex["zw"] = [vIndex["zp"][-1] + 1]
    vIndex["epsp"] = list(range(vIndex["zw"][-1] + 1, vIndex["zw"][-1] + n + 1))
    vIndex["espw"] = [vIndex["epsp"][-1] + 1]
    vIndex["delp"] = list(range(vIndex["espw"][-1] + 1, vIndex["espw"][-1] + n + 1))
    vIndex["delw"] = [vIndex["delp"][-1] + 1]
    vIndex["x"] = list(
        range(vIndex["delw"][-1] + 1, vIndex["delw"][-1] + reactions + 2)
    )
    nVars = vIndex["x"][-1] + 1

    for u in vIndex["u"]:
        problem.addVariable(
            index=u, prefix="u_", lb="infinity", ub="infinity", vtype="continuous"
        )

    for vp in vIndex["vp"]:
        problem.addVariable(
            index=vp, prefix="vp_", lb=0, ub="infinity", vtype="continuous"
        )

    for w in vIndex["w"]:
        problem.addVariable(
            index=w, prefix="w_", lb=0, ub="infinity", vtype="continuous"
        )

    for zp in vIndex["zp"]:
        problem.addVariable(index=zp, prefix="zp_", lb=0, ub=1, vtype="binary")

    for zw in vIndex["zw"]:
        problem.addVariable(index=zw, prefix="zw_", lb=0, ub=1, vtype="binary")

    for epsp in vIndex["epsp"]:
        problem.addVariable(
            index=epsp, prefix="epsp_", lb=0, ub="infinity", vtype="continuous"
        )

    for espw in vIndex["espw"]:
        problem.addVariable(index=espw, prefix="espw_", lb=0, ub=0, vtype="continuous")

    for delp in vIndex["delp"]:
        problem.addVariable(
            index=delp, prefix="delp_", lb=0, ub="infinity", vtype="continuous"
        )

    for delw in vIndex["delw"]:
        problem.addVariable(index=delw, prefix="delw_", lb=0, ub=0, vtype="continuous")

    for x in vIndex["x"][:-1]:
        problem.addVariable(
            index=x, prefix="x_", lb=0, ub="infinity", vtype="continuous"
        )
    problem.addVariable(
        index=vIndex["x"][-1], prefix="x_", lb=1000, ub="infinity", vtype="continuous"
    )

    vGroups["v"] = vIndex["vp"] + vIndex["w"]
    vGroups["z"] = vIndex["zp"] + vIndex["zw"]
    vGroups["eps"] = vIndex["epsp"] + vIndex["espw"]
    vGroups["del"] = vIndex["delp"] + vIndex["delw"]

    cIndex = {}

    cIndex["Ndual"] = list(range(0, reactions))
    cIndex["forceBioCons"] = [cIndex["Ndual"][-1] + 1]
    cIndex["forceKnockOut"] = [cIndex["forceBioCons"][-1] + 1]

    numberOfLinearCombinations = matrixS.shape[0] + matrixG.shape[0] + 1 + 1
    cIndex["linearCombination"] = list(
        range(
            cIndex["forceKnockOut"][-1] + 1,
            cIndex["forceKnockOut"][-1] + numberOfLinearCombinations,
        )
    )

    cIndex["forceLength"] = [cIndex["linearCombination"][-1] + 1]
    nCons = cIndex["forceLength"][-1] + 1

    A = scipy.sparse.lil_matrix((nCons, nVars))
    A[np.ix_(cIndex["Ndual"], vIndex["u"])] = matrixS.transpose()
    A[np.ix_(cIndex["Ndual"], vIndex["vp"])] = matrixG.transpose()
    A[np.ix_(cIndex["Ndual"], vIndex["w"])] = -t
    A[np.ix_(cIndex["forceBioCons"], vIndex["w"])] = -targetB
    A[np.ix_(cIndex["forceKnockOut"], vIndex["vp"])] = targetKOIndex

    auxS = scipy.sparse.hstack(
        [matrixS, scipy.sparse.lil_matrix(np.zeros((metabolites, 1)))]
    )
    auxK = scipy.sparse.hstack([matrixG, scipy.sparse.lil_matrix(np.zeros((n, 1)))])
    auxT = scipy.sparse.hstack([scipy.sparse.lil_matrix(-t), targetB])
    A[np.ix_(cIndex["linearCombination"], vIndex["x"])] = scipy.sparse.vstack(
        [auxS, auxK, auxT]
    )

    matrixEp1 = scipy.sparse.lil_matrix(
        (metabolites, (len(vIndex["vp"]) + len(vIndex["w"])))
    )
    matrixEp2 = scipy.sparse.eye((len(vIndex["vp"]) + len(vIndex["w"])), format="lil")
    auxEp = scipy.sparse.vstack([matrixEp1, -matrixEp2])
    A[np.ix_(cIndex["linearCombination"], vGroups["eps"])] = auxEp

    A[np.ix_(cIndex["linearCombination"], vGroups["del"])] = -auxEp

    rhs = np.zeros(nCons)
    rhs[cIndex["Ndual"]] = None
    rhs[cIndex["forceBioCons"][0]] = -cConstant
    rhs[cIndex["forceKnockOut"]] = None
    rhs[
        cIndex["linearCombination"][metabolites : len(cIndex["linearCombination"]) - 1]
    ] = targetKOIndex

    lhs = np.zeros(nCons)
    lhs[cIndex["Ndual"]] = 0
    lhs[cIndex["forceBioCons"][0]] = -1
    lhs[cIndex["forceKnockOut"]] = 10
    lhs[
        cIndex["linearCombination"][metabolites : len(cIndex["linearCombination"]) - 1]
    ] = targetKOIndex

    sense = np.array(["E"] * nCons)
    sense[cIndex["Ndual"]] = "G"
    sense[cIndex["forceBioCons"][0]] = "G"
    sense[cIndex["forceKnockOut"]] = "G"

    A = A.tocsr()
    problem.addConstraintsFromSparseMatrix(A, lhs, sense)

    varsForceBioCons = A[cIndex["forceBioCons"][0]].indices
    coefficients = A[cIndex["forceBioCons"][0]].data
    constraintName = f'c{cIndex["forceBioCons"][0]}'
    problem.modifyConstraint(
        constraintName=constraintName,
        variables=varsForceBioCons,
        coefficients=coefficients,
        sense="L",
        rhs=-cConstant,
    )
    # problem.addConstraint(name="forceBioCons_2",variables=varsForceBioCons, coefficients=coefficients, sense="L", rhs=-cConstant)

    vGroups["v"] = vIndex["vp"] + vIndex["w"]
    vGroups["z"] = vIndex["zp"] + vIndex["zw"]
    vGroups["eps"] = vIndex["epsp"] + vIndex["espw"]
    vGroups["del"] = vIndex["delp"] + vIndex["delw"]

    # Indicators
    # z = 1  -->  v >= alpha
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOn_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        sense="G",
        constant=ALPHA,
        activeWhen=1,
    )

    # z = 0  -->  v <= 0
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOff_",
        variables=vGroups["v"],
        indicatorVariables=vGroups["z"],
        sense="L",
        constant=0,
        activeWhen=0,
    )

    # z = 1  -->  epsilon <= 0
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOnEpsilon_",
        variables=vGroups["eps"],
        indicatorVariables=vGroups["z"],
        sense="L",
        constant=0,
        activeWhen=1,
    )

    # z = 0  -->  epsilon <= M
    problemMethods.buildIndicatorsIntoProblem(
        problem=problem,
        prefix="indicatorZOffEpsilon_",
        variables=vGroups["eps"],
        indicatorVariables=vGroups["z"],
        sense="L",
        constant=10,
        activeWhen=0,
    )

    vpVariableNames = {}
    for index, name in zip(vIndex["zp"], gDict.keys()):
        vpVariableNames[name] = index

    if numberOfRelationships > 0:
        subIndex = 0
        for key, values in relationships.items():
            vpContainer = vpVariableNames[key]
            for value in values:
                constraintName = f"relationship_constraint{subIndex}"
                vpContained = vpVariableNames[value]
                vars = [vpContainer, vpContained]
                coefs = [-1, 1]
                problem.addConstraint(constraintName, vars, coefs, "G", 0)
                subIndex += 1

    # Build objective
    objectiveVars = [
        vIndex["zp"][i]
        for i in range(0, len(possibleKOs))
    ]
    problem.addObjective(objectiveVars, list(numberNewGenesByKO.values()), sense="minimize")

    if forceLength:
        constraintName = "forceLength"
        vars = objectiveVars
        coefs = np.ones(len(objectiveVars))
        problem.addConstraint(constraintName, vars, coefs, "L", 1)

    return problem



from .Utilities import setSolver
from .Utilities import pd
from .ProblemInterpreters import cplexProblemInterpreter
from .ProblemInterpreters import gurobiProblemInterpreter
from .ProblemDefinitions import buildDictionaryMCSProblem
from .ProblemDefinitions import buildDictionaryMCSTargetedProblem
from .ProblemDefinitions import cobra
from .ProblemInterpreters import np
from .Utilities import beartype
from .ProblemDefinitions import List
from tqdm import tqdm

from .MinimalCutSetClass import MinimalCutSetProblem


def calculateMCS(
    cobraModel: cobra.Model,
    MAX_MCS: int,
    MAX_LENGTH_MCS: int,
    solver_params: dict = None,
    rxnSubset: List[str] = None,
    rxnKnockOut: List[str] = None,
    targetB=1e-3,
    timelimit=60,
    forceLength=True,
    numWorkers=0,
    verbose=0,
    useIndicators=True,
    solver: str = "gurobi",
    **kwargs,
):
    """Calculates the Minimal Cut Sets (MCS) of a metabolic network.
    Calculates the MCS of a metabolic network using a mixed integer linear programming (MILP) formulation.
    Among all reactions or with a given subset of reactions, the function finds the minimal set of reactions
    to stop biomass production.

    Required arguments:
        - **cobraModel:** A metabolic model in cobrapy format. See `cobra models <https://cobrapy.readthedocs.io/en/latest/io.html#Reading- and- Writing- Models>`_.
        - **MAX_MCS:** Maximum number of MCS to be found.
        - **MAX_LENGTH_MCS:** Maximum length of MCS to be found. (i.e. number of reactions in the MCS)

    Optional arguments:
        - **rxnToForce:** List of reactions to be considered in the MCS problem. The list is expected to be of reaction names. Default value is an empty list, which means that all reactions will be considered.

        - **rxnToKnockOut:** List of reactions that you want to be part of the MCS solutions. The list is expected to be of reaction names.
        
        - **targetB:** Desired activity level of the metabolic task to be disrupted. Default value is 1e- 3.

        - **timelimit:** Maximum time allowed for the solver to find a solution. Default value is 1e75.

        - **forceLength:** If True, the solver will find MCS of length 1, 2, 3, ..., MAX_LENGTH_MCS. If False, the solver will find MCS of length MAX_LENGTH_MCS. Default value is True.

        - **numWorkers:** Number of workers to be used for parallel processing. Default value is 0. The solvers cplex and gurobi already use parallel processing.

        - **verbose:** If 0, the solver will not print any information. Verbose  = 1, log will be printed into the console. Verbose = 2, log will be printed into the console and a log file will be created. Default value is 0.

        - **useIndicators:** If True, the solver will use binary variables to indicate if a reaction is part of the MCS.

        - **solver:** Solver to be used. Current compatibility includes cplex, gurobi, pulp and optlang_gurobi, optlang_cplex. Default value is cplex.

    """
    # Unpack the kwargs
    if kwargs:
        for key, value in kwargs.items():
            if key == "solver":
                solver = value
            elif key == "useIndicators":
                useIndicators = value
            elif key == "verbose":
                verbose = value
            elif key == "numWorkers":
                numWorkers = value
            elif key == "forceLength":
                forceLength = value
            elif key == "timelimit":
                timelimit = value
            elif key == "targetB":
                targetB = value
            elif key == "MAX_LENGTH_MCS":
                MAX_LENGTH_MCS = value
            elif key == "MAX_MCS":
                MAX_MCS = value
            else:
                raise ValueError("Unknown argument: " + key)

    gurobi_default_params = {
        "MIPFocus": 3,
        "TimeLimit": max(1, timelimit),
        "Threads": numWorkers,
        "PoolSolutions": 100,
        "PoolGap": 0.001,
        "PoolSearchMode": 1,
        "Presolve": 1,
    }

    cplex_default_params = {
        "mip.tolerances.integrality": 1e-5,
        "mip.strategy.rinsheur": 50,
        "emphasis.mip": 1,
        "timelimit": max(10, timelimit),
        "threads": numWorkers,
        "mip.limits.populate": 20,
        "mip.pool.relgap": 0.001,
    }


    if solver == "gurobi":
        params = gurobi_default_params
    elif solver == "cplex":
        params = cplex_default_params
    else:
        params = {}

    if solver_params is not None:
        for key, value in solver_params.items():
            if key in params:
                params[key] = value

    # Set the solver based on the user input, current compatibility includes cplex, gurobi, pulp and optlang_gurobi, optlang_cplex
    [interface, useIndicators] = setSolver(solver, useIndicators)

    # Constants
    integralityTolerance = 1e-5
    cConstant = 1e-3  # Used to activate w variable
    infinity = 1e30

    # Initialize the reactions to be knocked out, e.g. reactions that you want to be part of the MCS
    if rxnKnockOut is None:
        rxnKnockOut = []

    # Initialize the reactions to be forced, e.g. reactions that you want to force the flux through.
    if rxnSubset is None:
        rxnSubset = []

    # If there are reactions to be knocked out, then the problem is a targeted problem
    if rxnKnockOut:
        # Check that rxnToKnockOut is in the model and if rxnToForce is not empty check that rxnToKnockOut is in rxnToForce
        if not set(rxnSubset).issubset(set(cobraModel.reactions.list_attr("id"))):
            raise ValueError("One or more reactions in rxnToForce are not in the model")

        # merge rxnToForce and rxnKnockOut, or rxnToKnockOut with all reactions in the model
        if not rxnSubset:
            rxnSubset = cobraModel.reactions.list_attr("id")
        else:
            if not set(rxnKnockOut).issubset(set(rxnSubset)):
                raise ValueError(
                    "One or more reactions in rxnKnockOut are not in rxnToForce"
                )
        # Build the problem object
        problem = buildDictionaryMCSTargetedProblem(
            cobraModel=cobraModel,
            forceLength=forceLength,
            rxnSubset=rxnSubset,
            rxnKnockOut=rxnKnockOut,
        )
        # Set the solver
        problem.setSolver(solver)
        # Interpret the problem, depending on the solver each problem has to be interpreted differently into the solver's interface
        optimization = problem.interpretProblem(params)

        # Import the common problem methods
        problemMethods = MinimalCutSetProblem()

        # Get the objective variables, they are the variables that we are interested in
        obj = problem.getObjective()

        # Set the bounds for the solver, e.g. maximum number of solutions, maximum numbers in a solution, or if we want to force the step by step calculation etc.
        bounds = {
            "MAX_SOLUTION_LENGTH": MAX_LENGTH_MCS,
            "MAX_SOLUTIONS": MAX_MCS,
            "forceLength": False,
        }
        reactions = rxnSubset
        obj = obj["variables"]
        solutionVariables = {"indices": obj, "names": reactions}

        # Solve the problem iteratively
        solutionDict = problemMethods.iterateOverSolver(
            problem=problem,
            problemInterpreted=optimization,
            solver=solver,
            solutionVariables=solutionVariables,
            bounds=bounds,
        )      

        return solutionDict

    else:
        # If there are no specific reactions to be forced through all reactions in the model are included.
        if not rxnSubset:
            rxnSubset = cobraModel.reactions.list_attr("id")

        problem = buildDictionaryMCSProblem(
            cobraModel=cobraModel, forceLength=forceLength, rxnToForce=rxnSubset
        )

        problem.setSolver(solver)
        # Add the solver specific parameters

        optimization = problem.interpretProblem(params)
        problemMethods = MinimalCutSetProblem()
        obj = problem.getObjective()["variables"]
        reactions = rxnSubset
        solutionVariables = {"indices": obj, "names": reactions}
        bounds = {
            "MAX_SOLUTION_LENGTH": MAX_LENGTH_MCS,
            "MAX_SOLUTIONS": MAX_MCS,
            "forceLength": forceLength,
        }
        solutionDict = problemMethods.iterateOverSolver(
            problem=problem,
            problemInterpreted=optimization,
            solver=solver,
            solutionVariables=solutionVariables,
            bounds=bounds,
        )

        return solutionDict



# Generated by CodiumAI
from gMCSpy.Utilities import setSolver
import gMCSpy
import scipy
import numpy as np
import cplex
import gurobipy
try: 
    import pyscipopt
except ImportError:
    pass
import random

# Dependencies:
# pip install pytest-mock
import pytest

class TestSetsolver:

    # Tests that the function returns the correct interface and boolean value when the default solver is 'cplex' and useIndicators is True
    def test_default_solver_cplex_use_indicators_true(self, mocker):
        # Mock the import of cplex module
        mocker.patch('importlib.import_module', return_value=mocker.Mock())
    
        # Call the setSolver function with defaultSolver='cplex' and useIndicators=True
        result = setSolver('cplex', True)
    
        # Assert that the returned boolean value is True
        assert result[1] == True
    
        # Assert that the interface is cplex
        assert result[0] == cplex

    # Tests that the function returns the correct interface and boolean value when the default solver is 'gurobi' and useIndicators is True
    def test_default_solver_gurobi_use_indicators_true(self, mocker):
        # Mock the import of gurobipy module
        mocker.patch('importlib.import_module', return_value=mocker.Mock())
    
        # Call the setSolver function with defaultSolver='gurobi' and useIndicators=True
        result = setSolver('gurobi', True)

        # Assert that the returned boolean value is True
        assert result[1] == True
    
        # Assert that the interface is gurobi
        assert result[0] == gurobipy  

    # Tests that the function returns the correct interface and boolean value when the default solver is 'scip' and useIndicators is False
    def test_default_solver_scip_use_indicators_true(self, mocker):
        # Mock the import of pyscipopt module
        mocker.patch('importlib.import_module', return_value=mocker.Mock())
    
        # Call the setSolver function with defaultSolver='scip' and useIndicators=False
        #result = setSolver('scip', True)

        # Assert that the interface is pyscipopt
        #assert result[0] == pyscipopt    

        # Assert that the returned boolean value is False
        #assert result[1] == True
    


    # Tests that the function prints an error message when the default solver is not supported
    def test_default_solver_not_supported(self, capsys):
        # Call the setSolver function with defaultSolver='unsupported'
        setSolver('unsupported')
    
        # Capture the printed output
        captured = capsys.readouterr()
    
        # Assert that the error message is printed
        assert captured.out == "The default solver is not supported, please select (cplex, gurobi, scip)\n"

class TestCalculations:
    # Define the list of solvers to test
    solvers = ["cplex", "gurobi"]
    
    # test that the function returns the correct solution for the toy model using different solvers
    @pytest.mark.parametrize("solver", solvers)
    def test_GeneralityOfSolvers(self, solver):
        problem = gMCSpy.OptimizationProblem()
        problem.addVariable(0, 0, 'infinity', 'continuous')
        problem.addVariable(1, 0, 'infinity', 'continuous')
        problem.addVariable(2, 0, 'infinity', 'continuous')

        problem.addConstraint('c0',variables=[0,1,2], coefficients=[1,1,1], sense='L', rhs=100)
        problem.addConstraint('c1',variables=[0,1,2], coefficients=[10,4,5], sense='L', rhs=600)
        problem.addConstraint('c2',variables=[0,1,2], coefficients=[2,2,6], sense='L', rhs=300)

        problem.addObjective(variables=[0,1,2], coefficients=[10,6,4], sense='maximize')

        if solver == "gurobi":
            problem.setSolver(solver)
            [problemInterpreted, interface]= problem.interpretProblem()
            problemInterpreted.optimize()
            results = [
                (v.varName, v.X) for v in problemInterpreted.getVars() if abs(v.obj) > 1e-6
            ]
            objValue = round(problemInterpreted.objVal, 2)
            assert (objValue == 733.33) and (
                results
                == [("x_0", 33.333333333333336), ("x_1", 66.66666666666666), ("x_2", 0.0)]
            )
        elif solver == "cplex":
            problem.setSolver(solver)
            [problemInterpreted, interface]= problem.interpretProblem()
            problemInterpreted.solve()
            objVal = problemInterpreted.solution.get_objective_value()
            vars = problemInterpreted.solution.get_values()
            vars = [round(v, 2) for v in vars]
            assert (round(objVal, 2) == 733.33) and (vars == [33.33, 66.67, 0.0])
        elif solver == "scip":
            problem.setSolver(solver)
            [problemInterpreted, interface]= problem.interpretProblem()
            problemInterpreted.optimize()
            objVal = problemInterpreted.getObjVal()
            vars = problemInterpreted.getVars()
            vars = [round(problemInterpreted.getVal(v), 2) for v in vars]
            assert (round(objVal, 2) == 733.33) and (vars == [33.33, 66.67, 0.0])

    # Define the parameters to test
    FORCE_LENGTH_VALUES = [True, False]

    @pytest.mark.parametrize("solver", solvers)
    @pytest.mark.parametrize("force_length", FORCE_LENGTH_VALUES)
    def test_calculate_mcs_with_solver_and_force_length(self, solver, force_length):
        toy_model = gMCSpy.Utilities.createToyModel()
        mcs = gMCSpy.calculateMCS(toy_model, 1e6, 7, solver=solver,  forceLength=force_length)
        # Add assertions here to test the results with the specified solver
        expectedMCS = [['r1'], ['r7'], ['r8'], ['r2', 'r3'], ['r2', 'r4', 'r6'],  
                    ['r2', 'r4', 'r5']]
        falsePositive = False
        allResults = 0
        for information in mcs.items():
            solution = information[1]['solution']
            if solution in expectedMCS:
                allResults += 1
            else:
                falsePositive = True
        assert falsePositive == False
        assert allResults == len(expectedMCS)
        
    # Parametrize the tests with different solvers
    @pytest.mark.parametrize("solver", solvers)
    def test_calculate_mcs_rxn_knockout_multiple_times(self, solver):
        toy_model = gMCSpy.Utilities.createToyModel()
        for _ in range(3):  # Run the test three times
            possibleKOs = ['r1', 'r2', 'r3', 'r4', 'r5', 'r6', 'r7']
            random_element = random.choice(possibleKOs)
            mcs = gMCSpy.calculateMCS(toy_model, 1e6, 7, rxnKnockOut=[random_element], solver=solver)
            falsePositive = False
            for information in mcs.items():
                solution = information[1]['solution']
                if random_element not in solution:
                    falsePositive = True
            assert falsePositive == False
        
    # Parametrize the tests with different solvers
    @pytest.mark.parametrize("solver", solvers)
    def test_calculate_mcs_rxn_knockout_multiple(self, solver):
        toy_model = gMCSpy.Utilities.createToyModel()
        mcs = gMCSpy.calculateMCS(toy_model, 1e6, 7, solver=solver, rxnSubset=['r1','r2','r3','r4','r8'])
        expectedMCS = [['r1'], ['r8'], ['r2', 'r3']]
        falsePositive = False
        allResults = 0
        for information in mcs.items():
            solution = information[1]['solution']
            if solution in expectedMCS:
                allResults += 1
            else:
                falsePositive = True
        assert falsePositive == False
        assert allResults == len(expectedMCS)
        
    # Parametrize the tests with different solvers
    @pytest.mark.parametrize("solver", solvers)
    def test_calculate_gmcs_multiple(self, solver):
        toy_model = gMCSpy.Utilities.createToyModel()
        groundTruth = [['gene1'], ['gene2'], ['gene3'], ['gene6'], 
                       ['gene8'], ['gene9'], ['gene10']]
        solutions = gMCSpy.calculateGMCS.calculateGeneMCS(toy_model, solver=solver)
        count = 0
        for key, values in solutions.items():
            solution = values['solution']
            for gtSolution in groundTruth:
                if solution == frozenset(gtSolution):
                   count += 1
        assert count == len(groundTruth)


from cobra import Model, Reaction, Metabolite
from beartype import beartype
import os
from datetime import datetime
import pandas as pd
import re

def createToyModel():
    """Create a toy model for testing purposes.
    Creates a model with 9 reactions and 5 metabolites, the biomass reaction is **r8** based on the model in
    this `paper <https://academic.oup.com/bioinformatics/article/32/13/2001/1742871?login=true#394386436>`_.

    .. image:: ./figs/toymodel.jpeg
        :height: 350px
        :width: 400 px
        :align: center




    """
    model = Model("ToyModel")
    metaboliteA = Metabolite("m1", name="metaboliteA", compartment="c")
    metaboliteB = Metabolite("m2", name="metaboliteB", compartment="c")
    metaboliteC = Metabolite("m3", name="metaboliteC", compartment="c")
    metaboliteD = Metabolite("m4", name="metaboliteD", compartment="c")
    metaboliteE = Metabolite("m5", name="metaboliteE", compartment="c")

    reaction1 = Reaction("r1")
    reaction1.name = "reaction1"
    reaction1.subsystem = "subsystem1"
    reaction1.lower_bound = 0
    reaction1.upper_bound = 1000.0
    reaction1.add_metabolites({metaboliteA: 1.0})
    reaction1.gene_reaction_rule = "( gene1 and gene2 and gene3)"

    reaction2 = Reaction("r2")
    reaction2.name = "reaction2"
    reaction2.subsystem = "subsystem1"
    reaction2.lower_bound = -1000.0
    reaction2.upper_bound = 1000.0
    reaction2.add_metabolites({metaboliteA: -1.0, metaboliteD: 1.0})
    reaction2.gene_reaction_rule = "( gene1 or gene2 or gene4 or gene5 or gene6)"

    reaction3 = Reaction("r3")
    reaction3.name = "reaction3"
    reaction3.subsystem = "subsystem1"
    reaction3.lower_bound = 0.0
    reaction3.upper_bound = 1000.0
    reaction3.add_metabolites({metaboliteC: -1.0, metaboliteD: 1.0})
    reaction3.gene_reaction_rule = "( gene3 or (gene5 and gene6))"

    reaction4 = Reaction("r4")
    reaction4.name = "reaction4"
    reaction4.subsystem = "subsystem1"
    reaction4.lower_bound = 0.0
    reaction4.upper_bound = 1000.0
    reaction4.add_metabolites({metaboliteA: -1.0, metaboliteC: 1.0})
    reaction4.gene_reaction_rule = '((gene1 and gene2 and gene3 and gene4 and gene5) or (gene2 and gene3 and gene4 and gene6 and gene5) or (gene1 and gene2 and gene3 and gene7 and gene5) or (gene8 and gene2 and gene3 and gene4 and gene5) or (gene8 and gene2 and gene3 and gene7 and gene5) or (gene9 and gene2 and gene3 and gene7 and gene5) or (gene2 and gene3 and gene7 and gene6 and gene5) or (gene9 and gene2 and gene3 and gene4 and gene5)) and ((gene10 and gene11 and gene12 and gene13 and gene14 and gene15 and gene16 and gene17) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene14 and gene16 and gene17) or (gene10 and gene11 and gene12 and gene13 and gene14 and gene19 and gene15 and gene17) or (gene10 and gene11 and gene12 and gene13 and gene20 and gene14 and gene15 and gene16) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene20 and gene14 and gene16) or (gene10 and gene11 and gene12 and gene13 and gene15 and gene16 and gene21 and gene17) or (gene10 and gene11 and gene12 and gene13 and gene19 and gene15 and gene21 and gene17) or (gene10 and gene11 and gene12 and gene13 and gene20 and gene15 and gene16 and gene21) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene20 and gene22 and gene16) or (gene10 and gene11 and gene12 and gene13 and gene19 and gene15 and gene22 and gene17) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene20 and gene19 and gene21) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene19 and gene21 and gene17) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene20 and gene16 and gene21) or (gene10 and gene11 and gene12 and gene13 and gene15 and gene22 and gene16 and gene17) or (gene10 and gene11 and gene12 and gene13 and gene20 and gene15 and gene22 and gene16) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene16 and gene21 and gene17) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene22 and gene16 and gene17) or (gene10 and gene11 and gene12 and gene13 and gene20 and gene19 and gene15 and gene22) or (gene10 and gene11 and gene12 and gene13 and gene20 and gene19 and gene15 and gene21) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene20 and gene19 and gene22) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene19 and gene22 and gene17) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene14 and gene19 and gene17) or (gene10 and gene11 and gene12 and gene18 and gene13 and gene20 and gene14 and gene19) or (gene10 and gene11 and gene12 and gene13 and gene20 and gene14 and gene19 and gene15))'

    reaction5 = Reaction("r5")
    reaction5.name = "reaction5"
    reaction5.subsystem = "subsystem1"
    reaction5.lower_bound = 0.0
    reaction5.upper_bound = 1000.0
    reaction5.add_metabolites({metaboliteA: -1.0, metaboliteB: 1.0})
    reaction5.gene_reaction_rule = "( gene5 and gene2)"

    reaction6 = Reaction("r6")
    reaction6.name = "reaction6"
    reaction6.subsystem = "subsystem1"
    reaction6.lower_bound = 0.0
    reaction6.upper_bound = 1000.0
    reaction6.add_metabolites({metaboliteB: -1.0, metaboliteC: 1.0})
    reaction6.gene_reaction_rule = "( gene3 )"

    reaction7 = Reaction("r7")
    reaction7.name = "reaction7"
    reaction7.subsystem = "subsystem1"
    reaction7.lower_bound = 0.0
    reaction7.upper_bound = 1000.0
    reaction7.add_metabolites({metaboliteD: -1.0, metaboliteE: 1.0})
    reaction7.gene_reaction_rule = "( gene6 )"

    reaction8 = Reaction("r8")
    reaction8.name = "reaction8"
    reaction8.subsystem = "subsystem1"
    reaction8.lower_bound = 0.0
    reaction8.upper_bound = 1000.0
    reaction8.add_metabolites({metaboliteE: -1.0})
    reaction8.gene_reaction_rule = "( gene8 and gene9 and gene10)"

    reaction9 = Reaction("r9")
    reaction9.name = "reaction9"
    reaction9.subsystem = "subsystem1"
    reaction9.lower_bound = 0.0
    reaction9.upper_bound = 1000.0
    reaction9.add_metabolites({metaboliteB: -1.0})
    reaction9.gene_reaction_rule = "((gene2 and (gene5 or gene6)) or gene7)"

    model.add_reactions(
        [
            reaction1,
            reaction2,
            reaction3,
            reaction4,
            reaction5,
            reaction6,
            reaction7,
            reaction8,
            reaction9,
        ]
    )

    model.objective = "r8"

    return model

@beartype
def saveSolutions(
    problemDict: dict, solutionDict: dict, modelStruct: Model, filename: str, path: str
):
    """
    Saves the solutions to a csv file.

    .. warning::
        This function has been design to save the solution of our optimization problems, it may not work for other problems.

    :param problemDict: Dictionary containing the problem information.
    :param solutionDict: Dictionary containing the solution information.
    :param modelStruct: The cobra Model used to build the problem dictionary.
    :param filename: Name of the file to save the solutions.
    :param path: Path to save the file.


    """
    names = []
    orders = []
    reactionSets = []
    for key, data in solutionDict.items():
        vars = []
        for i in data:
            solPos = problemDict["variables"]["variableNames"].index(str(i))
            corresPondingVar = modelStruct.reactions[
                problemDict["variables"]["variableGroups"]["z"].index(solPos)
            ].id
            vars.append(corresPondingVar)
        orderNumber = key.split("order")[1].split("_")[0]
        names.append(key)
        orders.append(orderNumber)
        reactionSets.append(vars)

    df = pd.DataFrame({"solName": names, "order": orders, "Reactions": reactionSets})
    df.to_csv(path + "/" + filename + ".csv", index=False)

@beartype
def saveSolutionDict(dict: dict, filename: str, path: str):
    dataframe = pd.DataFrame.from_dict(dict, orient="columns")
    dataframe.to_csv(path + "/" + filename + ".csv", index=False)

def setSolver(defaultSolver, useIdicators: bool = True):
    """Set the solver to be used for the optimization problem.
    This function is used automatically set the correct interface based on the solver chosen.
    Current options are 'cplex', 'gurobi' and 'scip'.

    :param defaultSolver: The solver to be used for the optimization problem. Must be one of the options above.
    :param indicator: If True, the solver will be set to use indicators. If False, the solver will be set to not use indicators.
        The function also checks if the solver has the capability to use indicators. If the solver does not have the capability
        it returns turns the indicators off and the Big M method is implemented.

    :return:
        - The interface to be used for the optimization problem.
        - A boolean indicating if the solver will use indicators or not.
    """
    if defaultSolver == "cplex":
        try:
            import cplex

            return [cplex, True]
        except ImportError :
            print(
                "cplex could not be loaded, check if it is installed and the path is correct"
            )
    elif defaultSolver == "gurobi":
        try:
            import gurobipy

            return [gurobipy, True]
        except ImportError :
            print(
                "gurobi could not be loaded, check if it is installed and the path is correct"
            )

    elif defaultSolver == "scip":
        try: 
            import pyscipopt

            return [pyscipopt, True]
        except ImportError :
            print(
                "pyscipopt could not be loaded, check if it is installed and the path is correct"
            )
    else:
        print(
            "The default solver is not supported, please select (cplex, gurobi, scip)"
        )

def createLog(folderLogPath: str = "logs/log"):
    """Creates a log file with the current date and time, and returns the filename.
    This functions also creates the folder to save the log file if it does not exist.
    It is used to save the log of the optimization process depending on the verbosity level."""
    filename = folderLogPath
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    print(filename)
    return filename

def loadSolutions(path: str = None):
    """Load the solutions generated from the calculation of gMCSs.

    :param path: Path to the folder with the solutions, if no path is given the most recent calculation in the logs folder is used.

    :return: A dictionary with the solutions.
    """
    if path is None:
        path = 'logs'
        # Get a list of all directories in the given path
        directories = [d for d in os.listdir(path) if os.path.isdir(os.path.join(path, d))]
        
        # Sort the directories by their modification time in descending order
        directories.sort(key=lambda x: os.path.getmtime(os.path.join(path, x)), reverse=True)
        
        if directories:
            # The most recent folder is the first one in the sorted list
            most_recent = directories[0]
            complePath = path + "/" + most_recent + "/solutions/solutions.log"
            return read_sol_log(complePath)
            
        else:
            raise Warning("No solutions found in the given path.")
    else:
        return read_sol_log(path)  

def read_sol_log(logFile):
    # Read the log file
    with open(logFile, "r") as file:
        # First line are headers
        lines = file.readlines()[1:]
        solDict = {}
        for line in lines:
            line = line.strip()
            # Split the line by the separator
            lineList = line.split(":")
            # Get the genes
            geneList = add_quotes_inside_brackets(lineList[0])
            geneList = eval(geneList)
            geneList = frozenset(geneList)
            # Get the reactions in the solution
            reactionList = eval(add_quotes_inside_brackets(lineList[1]))
            solDict[geneList] = reactionList
        
        return solDict
        



def add_quotes_inside_brackets(input_string):
    # Use regular expressions to find text inside square brackets
    pattern = r'\[([^]]*)\]'
    
    # Define a function to add quotes around the found text
    def add_quotes(match):
        # Split the text inside brackets by ',' and add quotes to each element
        elements = match.group(1).split(',')
        quoted_elements = [f'"{element.strip()}"' for element in elements]
        return '[' + ','.join(quoted_elements) + ']'
    
    # Use re.sub() to apply the function to the input string
    result = re.sub(pattern, add_quotes, input_string)
    
    return result


from .ProblemDefinitions import cobra
from .ProblemDefinitions import buildDictionaryFBAWithDeletions
from .MinimalCutSetClass import logging
from .MinimalCutSetClass import tqdm
from. MinimalCutSetClass import prepareModelNutrientGeneMCS
from .Utilities import datetime
from .Utilities import os

import time

from joblib import Parallel, delayed


def checkGMCS(
    geneSolution,
    model: cobra.Model,
    isoFormSeparator=None,
):
    problem = buildDictionaryFBAWithDeletions(model, geneSolution, isoFormSeparator)
    problem.setSolver("gurobi")
    gurobi_default_params = {
        "Threads": 1,
        "OutputFlag": 0
        }
    [problemInterpreted, _] = problem.interpretProblem(parameters=gurobi_default_params)
    problemInterpreted.optimize()
    try:
        cutSetValue = problemInterpreted.ObjVal
    except:
        checkDict = {
        "isGMCS": 'infesible',
        "solution": geneSolution,
        "isCutSet": 'infesible',
        "cutSetValue": 'infesible',
        "isMinimal": 'infesible',
        "setGeneTested": 'infesible',
        "objVals": 'infesible',
        }
        return checkDict

    if problemInterpreted.ObjVal > 1e-6:
        isCutSet = False
    else:
        isCutSet = True

    isMinimal = []
    setGeneTested = []
    objVals = []
    for i in range(len(geneSolution)):
        ko = list(geneSolution)

        if len(ko) > 1:
            ko.remove(ko[i])
            problem = buildDictionaryFBAWithDeletions(model, ko, isoFormSeparator)
            problem.setSolver("gurobi")
            [problemInterpreted, interface] = problem.interpretProblem(parameters=gurobi_default_params)
            problemInterpreted.optimize()
            objVals.append(problemInterpreted.ObjVal)
            if problemInterpreted.ObjVal > 0:
                isMinimal.append(True)
            else:
                isMinimal.append(False)

        elif len(ko) == 1 and isCutSet:
            isMinimal.append(True)
            objVals.append(0)

        setGeneTested.append(ko)

    areAllMinimal = all(isMinimal)
    both = [isCutSet, areAllMinimal]
    isGMCS = all(both)
    checkDict = {
        "isGMCS": isGMCS,
        "solution": geneSolution,
        "isCutSet": isCutSet,
        "cutSetValue": cutSetValue,
        "isMinimal": isMinimal,
        "setGeneTested": setGeneTested,
        "objVals": objVals,
    }
    return checkDict


# Build parrallel function to check all solutions in a list

def checkGMCSParallel(knockoutList, model, workers, isoFormSeparator=None, log_location=None, isNutrient=False, handler=None):
    
    if len(knockoutList) == 0:
        print("No solutions to check")
        return
    
    if handler:
        loggerMain = logging.getLogger(handler)
    else:
        loggerMain = None
    
    timeStart = time.time()
    
    if isNutrient:
        model = prepareModelNutrientGeneMCS(model, [])
        
    # Set the location and filename for the log file
    if log_location is None:
        folderLogPath = "checks/check_"
        log_location = folderLogPath + model.id + datetime.now().strftime("%Y_%m_%d-%I_%M_%S") + ".txt"
        os.makedirs(os.path.dirname(log_location), exist_ok=True)
        print('saving check results to: ' + log_location)
    else:
        log_location = log_location + "checkedSolutions.csv"
        os.makedirs(os.path.dirname(log_location), exist_ok=True)
        print('saving check results to: ' + log_location)
    # Create a logger
    logger = logging.getLogger('checks')
    logger.setLevel(logging.DEBUG)

    # Create a file handler
    file_handler = logging.FileHandler(log_location)
    file_handler.setLevel(logging.DEBUG)

    # Customize the log message format
    formatter = logging.Formatter('%(message)s')  # Only includes the log message

    # Set the formatter for the file handler
    file_handler.setFormatter(formatter)

    # Add the file handler to the logger
    logger.addHandler(file_handler)

    # Write the first line to the log
    logger.debug('model,isGMCS,completeDict')
    
    # zip al parameters into a list repeating model and isoFormSeparator
    listModels = [model] * len(knockoutList)
    listIsoFormSeparator = [isoFormSeparator] * len(knockoutList)
    zipppedParameters = zip(knockoutList, listModels, listIsoFormSeparator)

    # Redirect stdout to a null device
    with open(os.devnull, "w") as devnull:
        original_stdout = os.dup(1)  # Save the original stdout
        os.dup2(devnull.fileno(), 1)  # Replace stdout with the null device

    results_iter = Parallel(n_jobs=workers)(delayed(checkGMCSChild)(zippedParameters) for zippedParameters in tqdm(zipppedParameters, total=len(knockoutList), 
                                                                                                              desc="Checking GMCSs", unit="GMCS"))

    # Restore the original stdout
    os.dup2(original_stdout, 1)

    percentageOfTrue = []
    for result in results_iter:
        logger.debug(f"{model.id},{result['isGMCS']},{result}")
        percentageOfTrue.append(result['isGMCS'])
    
    endTime = time.time()
    timeElapsed = endTime - timeStart
    if loggerMain:
        loggerMain.info(f"checks: {timeElapsed}")
    print('Percentage of True GMCS: ' +  str((sum(percentageOfTrue)/len(percentageOfTrue))*100))

    logger.removeHandler(file_handler)
    file_handler.close()
    

def checkGMCSChild(zippedParameters):
    knockout = zippedParameters[0]
    model = zippedParameters[1]
    isoFormSeparator = zippedParameters[2]  
    checkDict = checkGMCS(knockout, model, isoFormSeparator)
    return checkDict




from .ProblemDefinitions import cobra
from .ProblemInterpreters import np


from .ProblemInterpreters import beartype
from .ProblemDefinitions import List
from .ProblemDefinitions import scipy
from .ProblemDefinitions import buildDictionaryMCSGeneProblem
from .ProblemDefinitions import buildDictionaryMCSGeneTargetedProblem

from .Validations import checkGMCSParallel

from .calculateMCS import calculateMCS

import re
from ast import And, BoolOp, Or, Name

from collections import defaultdict
import warnings
import time

from .MinimalCutSetClass import MinimalCutSetProblem
from .MinimalCutSetClass import uuid
from .MinimalCutSetClass import tqdm
from .MinimalCutSetClass import logging
from .MinimalCutSetClass import prepareModelNutrientGeneMCS

from datetime import datetime
import pandas as pd

from .Utilities import createLog


def calculateGeneMCS(cobraModel, **kwargs):
    '''
    **Calculate the genetic minimal cut sets of a model**
    Calculates the gMCS of a metabolic network using a mixed integer linear programming (MILP) formulation.
    Among all genes or with a given subset of genes, the function finds the minimal set of genes
    to stop biomass production.

    Required arguments:
        - **cobraModel:** A metabolic model in cobrapy format. See `cobra models <https://cobrapy.readthedocs.io/en/latest/io.html#Reading- and- Writing- Models>`_.

    Optional arguments:
        - **Solver:** Solver to be used. Current compatibility includes cplex, gurobi and SCIP. Default value is gurobi.
        - **maxKOLength :** Maximum length of gMCS to be found. (i.e. number of genes in the gMCS). Default value is 3
        - **maxNumberGMCS :** Maximum number of gMCS to be found. Default value is 1e6
        - **verbose:** If 0, the solver will not print any information. Verbose  = 1, log will be printed into the console. Verbose = 2, log will be printed into the console and a log file will be created. Default value is 0.
        - **earlyStop :** If True if optimality returned by the solver the solver will stop without running a second time. Default value is True.
        - **removeRelated:** If True, the solver will remove the gMCS that are a subset of another gMCS from the solution space. Default value is True.
        - **targetKOs:** List of genes that you want to be part of the gMCS solutions. The list is expected to be of gene names. Default value is an empty list, which means that all genes will be considered.
        - **geneSubset:** List of genes to be considered in the gMCS problem. The list is expected to be of gene names. Default value is an empty list, which means that all genes will be considered.
        - **isNutrient:** If True, the solver will consider the exchange reactions as nutrients, assigning artificial genes to exchange reactions to evaluate what metabolites and genes are essential. Default value is False.
        - **exchangeReactions:** List of exchange reactions to be considered as nutrients. The list is expected to be of reaction names. Default value is None.
        - **saveGMatrix:** If True, the solver will save the G matrix in a csv file. Default value is False.
        - **checkSolutions:** If True, the solver will check if the solutions are valid. Default value is False.
        - **saveSolutions:** If True, the solver will save the solutions in a csv file. Default value is True.
        - **timelimit:** Maximum time allowed for the solver to find a solution. Default value is 1e75.
        - **forceLength:** If True, the solver will find MCS of length 1, 2, 3, ..., MAX_LENGTH_MCS. If False, the solver will find MCS of length MAX_LENGTH_MCS. Default value is True for cplex and false for gurobi.
        - **numWorkers:** Number of workers to be used for parallel processing. Default value is 0. The solvers cplex and gurobi already use parallel processing.      
        - **isoformSeparator:** If not None, the isoforms of the genes will merge into one gene. Default value is None.
    '''
    kwargs["solver"] = kwargs.get("solver", "gurobi")
    name = (
        "gMCSpy_"
        + cobraModel.id
        + "_"
        + kwargs["solver"]
        + "_"
        + datetime.now().strftime("%Y%m%d-%H%M%S")
    )
    
    path = "logs/" + name + "/process/"
    createLog(path)
    path = "logs/" + name
    name = ("process")
    handler = name + ".log"
    logging = initiateLogger(name + ".log", path + "/process/" + name + ".log")
    allTime = time.time()

    maxKOLength = kwargs.get('maxKOLength', 3)
    maxNumberGMCS = kwargs.get('maxNumberGMCS', 1e6)
    verbose = kwargs.get('verbose', 0)
    earlyStop = kwargs.get("earlyStop", True)
    removeRelated = kwargs.get("removeRelated", False)
    isoformSeparator = kwargs.get("isoformSeparator", None)
    targetKOs = kwargs.get("targetKOs", None)
    geneSubset = kwargs.get("geneSubset", None)
    isNutrient = kwargs.get("isNutrient", False)
    exchangeReactions = kwargs.get("exchangeReactions", None)
    saveGMatrix = kwargs.get("saveGMatrix", False)
    checkSolutions = kwargs.get("checkSolutions", False)
    saveSolutions =  kwargs.get('saveSolutions', True)
    timeLimit = kwargs.get("timeLimit", 1e4)
    solver = kwargs["solver"]
    
    if kwargs.get("forceLength") is None:
        if solver == "gurobi":
            forceLength = False
        elif solver == "cplex":
            forceLength = True
        elif solver == "scip":
            forceLength = True   
    else:
        forceLength = kwargs.get("forceLength", False)
    
    kwargs.pop("maxKOLength", None)
    kwargs.pop("maxNumberGMCS", None)
    kwargs.pop("verbose", None)
    kwargs.pop("earlyStop", None)
    kwargs.pop("removeRelated", None)
    kwargs.pop("isoformSeparator", None)
    kwargs.pop("targetKOs", None)
    kwargs.pop("geneSubset", None)
    kwargs.pop("isNutrient", None)
    kwargs.pop("exchangeReactions", None)
    kwargs.pop("saveGMatrix", None)
    kwargs.pop("checkMCS", None)
    kwargs.pop('saveSolutions', None)
    kwargs.pop("forceLength", None)
    kwargs.pop("timeLimit", None)

    numWorkers = kwargs.get("numWorkers", 0)
    

    
    if isNutrient:
        cobraModel = prepareModelNutrientGeneMCS(cobraModel, exchangeReactions)

    logging.info(
        f"model: {cobraModel.id}, maxKOLength: {maxKOLength}, maxNumberGMCS: {maxNumberGMCS}, verbose: {verbose}, forceLength: {forceLength}, earlyStop: {earlyStop}, removeRelated: {removeRelated}, solver: {kwargs['solver']}"
    )
    startTime = time.time()
    if isoformSeparator is not None:
        gObject = buildGMatrix(
            cobraModel,
            maxKOLength,
            isoformSeparator=isoformSeparator,
            verbose=0,
            **kwargs,
        )
    else:
        gObject = buildGMatrix(cobraModel, maxKOLength, verbose=0, **kwargs)
    endTime = time.time()
    logging.info("GMatrix: " + str(endTime - startTime))
    
    gDict = gObject["gDict"]
    gMatrix = gObject["gMatrix"]
    relationships = gObject["relationships"]
    numberNewGenesByKO = gObject["numberNewGenesByKO"]
    
    # Filter the G matrix to keep only genes that are in the geneSubset
    if geneSubset:
        #Make a set of the geneSubset
        geneSet = set(geneSubset)
        # Extract all possible interventions
        gDictKeys = list(gDict.keys())
        # Remove the interventions that are not in the geneSubset
        for key in gDictKeys:
            if not key.issubset(geneSet):
                gDict.pop(key)

        gMatrix = __createSparseMatrix(gDict, cobraModel.reactions)
        [relationships, numberNewGenesByKO] = __relatedRows(gDict, mergeIsforms(isoformSeparator))
   
    if saveGMatrix:        
        gMatrix_Strings = []
        for ko, reactions  in gDict.items():
            string_ko = "@@".join(list(ko))
            string_reactions = "@@".join(cobraModel.reactions[x].id for x in reactions)
            row = string_ko + " -> " + string_reactions
            gMatrix_Strings.append(row)
        filename = f"Gmat_python_{cobraModel.id}_{solver}.csv"
        gPath = path + "/GMatrix/" 
        createLog(gPath)
        completePath = gPath + filename
        df = pd.DataFrame({'gMatrix': gMatrix_Strings})
        df.to_csv(completePath)


    # Initialize the genes to be knocked out, e.g. genes that you want to be part of the gMCS
    if targetKOs is None:
        targetKOs = []

    # Make sure that targetKOs is a list
    if not isinstance(targetKOs, list):
        raise ValueError("targetKOs must be a list")

    # Initialize the subset of genes to be studied , e.g. the space of genes to calculate the gMCS
    if geneSubset is None:
        geneSubset = []

    # Make sure that geneSubset is a list
    if not isinstance(geneSubset, list):
        raise ValueError("geneSubset must be a list")

    # List of possible interventions
    gDictKeys = list(gDict.keys())

    # If there are reactions to be knocked out, then the problem is a targeted problem
    if targetKOs:
        # Transform each element of the targetKOs into a frozenset to be able to compare it with the keys of the gDict
        targetKOs = [frozenset([gene]) for gene in targetKOs]
        # Check that the targetKOs is in the model
        for gene in targetKOs:
            isInModel = []
            for intervention in gDictKeys:
                if gene.issubset(intervention):
                    isInModel.append(intervention)
                    break
            if not isInModel:
                raise ValueError("One or more genes in targetKOs are not in the model")

        # Check that geneSubset is not an empty list
        if geneSubset:
            # Transform each element of the geneSubset into a frozenset to be able to compare it with the keys of the gDict
            geneSubset = [frozenset([gene]) for gene in geneSubset]
            # Check that the geneSubset is in the model
            for gene in geneSubset:
                isInModel = []
                for intervention in gDictKeys:
                    if gene.issubset(intervention):
                        isInModel.append(intervention)
                        break
                if not isInModel:
                    raise ValueError(
                        "One or more genes in geneSubset are not in the model"
                    )
        else:
            geneSubset = []
            for intervention in gDictKeys:
                if len(intervention) == 1:
                    geneSubset.append(intervention)

        geneSubset = list(set(geneSubset).union(set(targetKOs)))

        # Build the problem object
        startTime = time.time()
        problem = buildDictionaryMCSGeneTargetedProblem(
            cobraModel,
            gDict,
            gMatrix,
            relationships=relationships,
            geneSubset=geneSubset,
            targetKOs=targetKOs,
            maxKOLength=maxKOLength,
            forceLength=forceLength,
            numberNewGenesByKO=numberNewGenesByKO,
            verbose=verbose,
        )
        endTime = time.time()
        logging.info(
            "Time taken to build problem: " + str(endTime - startTime) + " seconds"
        )

        gurobi_default_params = {
            "MIPFocus": 1,
            "TimeLimit": max(1, timeLimit),
            "Threads": numWorkers,
            "PoolSolutions": 10,
            "PoolGap": 0.1,
            "PoolSearchMode": 1,
            "Cutoff": maxKOLength,
            "Presolve": 1,
            "PreSOS1Encoding": 2,
            "Cuts": 2,
        }

    else:
        startTime = time.time()
        problem = buildDictionaryMCSGeneProblem(
            cobraModel,
            gDict,
            gMatrix,
            relationships=relationships,
            maxKOLength=maxKOLength,
            forceLength=forceLength,
            numberNewGenesByKO=numberNewGenesByKO,
            verbose=verbose,
            logPath=path
        )
        endTime = time.time()
        logging.info("BuildProblem: " + str(endTime - startTime))

            
        gurobi_default_params = {
            "MIPFocus": 3,
            "TimeLimit": max(1, timeLimit),
            "Threads": numWorkers,
            "PoolSolutions": 10000,
            "PoolGap": 0.1,
            "PoolSearchMode": 1,
            "Cutoff": maxKOLength,
            "Presolve": 1,
            "PreSOS1Encoding": 2,
            "Cuts": 2,
            #"OptimalityTol": 1e-9,
        }

        if forceLength:
            gurobi_default_params['MIPFocus'] = 3
            gurobi_default_params['TimeLimit'] = 500
            
        '''
            gurobi_default_params["MIPFocus"] = 1
        '''
    problem.setCharacteristicParameter(maxKOLength)
    problem.setModelName(cobraModel.id)
    problem.setLogPath(path)
    # Set the solver
    
    problem.setSolver(solver)

    cplex_default_params = {
        "mip.tolerances.integrality": 1e-5,
        "emphasis.mip": 4,
        "timelimit": max(10, timeLimit),
        "threads": numWorkers,
        "preprocessing.presolve": 1,
        "mip.limits.populate": 1000,
        "mip.pool.relgap": 0.1,
        "mip.tolerances.uppercutoff": maxKOLength,
        "mip.pool.intensity": 3,
        "mip.strategy.probe": 2,
        "mip.strategy.variableselect": 4,
        "preprocessing.sos1reform": 1,
    }


    if solver == "gurobi":
        params = gurobi_default_params
    elif solver == "cplex":
        params = cplex_default_params
    else:
        params = {}
    # Interpret the problem, depending on the solver each problem has to be interpreted differently into the solver's interface
    optimization = problem.interpretProblem(verbose=2, parameters=params)

    # Import the common problem methods
    problemMethods = MinimalCutSetProblem()

    # Get the objective variables, they are the variables that we are interested in
    knockouts = list(gDict.keys())
    zpVariables = [
        key for key, data in problem.getVariables().items() if "zp_" in data["name"]
    ]
    solutionVariables = {"indices": zpVariables, "names": knockouts}
    # Set the bounds for the solver, e.g. maximum number of solutions, maximum numbers in a solution, or if we want to force the step by step calculation etc.
    bounds = {
        "MAX_SOLUTION_LENGTH": maxKOLength,
        "MAX_SOLUTIONS": maxNumberGMCS,
        "forceLength": forceLength,
    }

    # Solve the problem iteratively
    solutionDict = problemMethods.iterateOverSolver(
        problem=problem,
        problemInterpreted=optimization,
        solver=solver,
        solutionVariables=solutionVariables,
        bounds=bounds,
        mergeSolutions=True,
        earlyStop=earlyStop,
        removeRelated=removeRelated,
        verbose=2,
        handler=handler,
        iterationProgress=True,
    )
    allEndTime = time.time()

    logging.info("totalTime: " + str(allEndTime - allTime))
    
    if saveSolutions:
        solutionPath = path + '/solutions/'
        solutionFileName = 'solutions.log'
        createLog(solutionPath)
        loggingSolutions = initiateLogger(solutionFileName, solutionPath + solutionFileName)
        loggingSolutions.info('gene,reactions')
        geneSolutionList = []
        
        for order, ko in solutionDict.items():
            reactions_log_list = []
            for gene in ko['solution']:
                reactions = gDict[frozenset([gene])]
                reactions_log_list.append([cobraModel.reactions[x].id for x in reactions])
            reactions = ','.join(list(set().union(*reactions_log_list)))
            ko_log = ','.join(list(ko['solution']))         
            loggingSolutions.info(f'[{ko_log}]:[{reactions}]')

        solutionHandler = loggingSolutions.handlers[0]
        loggingSolutions.removeHandler(solutionHandler)
        solutionHandler.close()
               
    if checkSolutions:
        geneSolutionList = []
        for key, items in solutionDict.items():
            geneSolutionList.append(items['solution'])
        
        checkGMCSParallel(geneSolutionList, cobraModel, 16, isoFormSeparator=isoformSeparator, log_location=path + "/checks/", handler=handler)

    handler = logging.handlers[0]
    logging.removeHandler(handler)
    handler.close()
    return solutionDict




def buildGMatrix(
    cobraModel: cobra.core.Model,
    maxKOLength: int = 1e6,
    name: str = None,
    isoformSeparator: str = None,
    verbose: int = 0,
    **kwargs,
):
    """

    **Build the G matrix, G dict and the relationships between perturbations**

    The G matrix is a matrix with the perturbations of the genes in the model, that render a reaction infeasible.
    Each row of the matrix corresponds to a perturbation, and each column corresponds to a reaction.
    This matrix is used to calculate the genetic minimal cut sets of the model.

    :param cobraModel: The cobra model that represents the metabolism of an organism.
    :param name: The name of the model. If None, the id of the model will be used.
    :param separateIsoform: If not None, the isoforms of the genes will merge into one gene.
    :param verbose: The level of verbosity of the function.

    :return: A List with the G matrix, the G dictionary and the relationships between perturbations.


    """
    
    maxKOLength = kwargs.get('maxKOLength', 3)
    dictManager = mergeIsforms(isoformSeparator)
        
    modelReactions = [reaction.id for reaction in cobraModel.reactions]
    # create a dictionary with the gpr rules associated to each reaction to get a unique list of gprs
    gprDict = getGPRDict(cobraModel)
    gDict = {}
    
    # Analyze the GPRs to find the perturbations that stop the reactions
    for key, value in tqdm(gprDict.items()):
            calculate_GPR_MCS(gDict, value['gpr'], value['reactions'], dictManager)


    # Filter out the interventions that are larger than the maxKOLength
    filteredGDict = gDict.copy()
    for key in gDict.keys():
        if len(key) > maxKOLength:
            filteredGDict.pop(key)
            
    newGDict = transformReactionIdsIntoIndexes(filteredGDict, modelReactions)
            
    simpG = __simplifyGMatrix(newGDict, dictManager)

    [relationships, numberNewGenesByKO] = __relatedRows(simpG, dictManager)
    
    gMatrix = __createSparseMatrix(simpG, modelReactions)

    # scipy.sparse.save_npz(cobraModelName + '_gMatrix.npz', gMatrix)
    gObject = {
        "gMatrix": gMatrix,
        "gDict": simpG,
        "relationships": relationships,
        "numberNewGenesByKO": numberNewGenesByKO,
    }
    return gObject

def calculate_GPR_MCS(gDict, gpr, reactions, addToDict):
    # get the genes from the gpr rule
    stringGPR = gpr.to_string()
    genes = [gene for gene in gpr.genes]
    if len(genes) == 1:
        key=frozenset(genes)
        [addToDict(gDict, key, reaction) for reaction in reactions]
    # Only and
    elif bool(re.search(' and ', stringGPR) and not re.search(' or ', stringGPR)):
        # Only one gene needs to be KO to stop the reaction
        for gene in genes:
            key=frozenset([gene])
            [addToDict(gDict, key, reaction) for reaction in reactions]

    # Only or
    elif bool(re.search(' or ', stringGPR) and not re.search(' and ', stringGPR)):
        # All genes need to be KO to stop the reaction
        key=frozenset(genes)
        [addToDict(gDict, key, reaction) for reaction in reactions]  
    
    # And and or
    else:
        treeGenes, model = parseGPRToModel(stringGPR, 'gpr', None)
        solutionDict = calculateMCS(model, MAX_LENGTH_MCS=5, MAX_MCS=1e6, rxnSubset=list(treeGenes))
        for _, data in solutionDict.items():
            key = frozenset(data["solution"])
            [addToDict(gDict, key, reaction) for reaction in reactions]
                   
def getGPRDict(model):
    gpr_dict = {}
    for reaction in model.reactions:
        gpr = reaction.gpr.to_string()
        if gpr in gpr_dict:
            gpr_dict[gpr]['reactions'].append(reaction.id)
        elif gpr == '':
            pass
        else:
            gpr_dict[gpr] = {'reactions': [reaction.id], 'genes': [gene.id for gene in reaction.genes], 'gpr': reaction.gpr}
    return gpr_dict

def __simplifyGMatrix(gMatrixDict: defaultdict, __addToDict) -> defaultdict:
    """**Simplify the G matrix**

    Analyze the G matrix to find the rows that are related to each other.
    If a common gene is more perturbations, then the perturbations are related to each other.
    And should be an individual perturbation.

    :param gMatrixDict: The dictionary with the perturbations.

    :return: The simplified G matrix.
    """
    
    # Empty dictionary to store the simplified G matrix
    simpG = gMatrixDict.copy()
    
    # Get the interventions compossed of one gene
    singleGeneInterventions = [gene for gene in gMatrixDict.keys() if len(gene) == 1]
    
    # Get the interventions compossed of more than one gene
    multipleGeneInterventions = [gene for gene in gMatrixDict.keys() if len(gene) > 1]
    
    # Intersect each intervention with the other interventions to find common elements
    for interventionA in multipleGeneInterventions:
        for interventionB in multipleGeneInterventions:
            if interventionA != interventionB:
                intersection = interventionA.intersection(interventionB)
                if intersection:
                    # If the intersection is not empty, then the interventions are related
                    # Check if the intersection is already a single gene intervention
                    if len(intersection) == 1:
                       if intersection not in singleGeneInterventions and intersection not in gMatrixDict.keys():
                           # Intersection must not be a key of the dictionary
                            if intersection not in simpG.keys():
                                 # Add the intersection as a key of the dictionary
                                __addToDict(simpG, intersection, [])
                                simpG[intersection] = []
                    else:
                        for element in intersection:
                            if element not in singleGeneInterventions and element not in gMatrixDict.keys():
                                # Intersection must not be a key of the dictionary
                                if intersection not in simpG.keys():
                                     # Add the intersection as a key of the dictionary
                                    __addToDict(simpG, intersection, [])
                                    simpG[intersection] = []
                                

    return simpG

def __createSparseMatrix(gMatrixDict, modelReactions):
    """**Build the G matrix**

    Create sparse matrix from the dictionary with the perturbations.

    :param gMatrixDict: The dictionary with the perturbations.
    :param modelReactions: The reactions of the model.

    :return: The G matrix.
    """
    gMatrix = scipy.sparse.lil_matrix((len(gMatrixDict), len(modelReactions)))
    for i, items in enumerate(gMatrixDict.items()):
        gMatrix[i, list(items[1])] = 1
    gMatrix = gMatrix.tocsc()
    return gMatrix

def __relatedRows(gMatrixDict: defaultdict, __addToDict) -> List:
    """**Find the rows that are related to each other**

    If a row is a subset of another row,
    then that means that when the perturbation of the superset is activated, the subset is also affected.
    i. e. gene A and gene B inactivates reaction1, and gene A inactivates reaction2. Knocking out gene A and gene B will affect both reactions.

    :param gMatrixDict: The dictionary with the perturbations.

    :return: A list with the rows that are related to each other.
    """
    relationships = defaultdict(list)
    numberNewGenesByKO = defaultdict(int)
    sortedDict = sorted(gMatrixDict.items(), key=lambda x: len(x[0]))
    for key, value in gMatrixDict.items():
        numberNewGenesByKO[key] = len(key)
        if len(key) < 2:
            continue
        keysConstained = []
        for key2, value2 in sortedDict:
            if len(key2) >= len(key):
                break
            if key2.issubset(key):
                keysConstained.append(key2)
                value.extend(value2)
                __addToDict(relationships, key, key2)
        if len(keysConstained) > 0:
            key2 = frozenset.union(*keysConstained)
            numberNewGenesByKO[key] = len(key) - len(key2)
        gMatrixDict[key] = list(set(value))
    return [relationships, numberNewGenesByKO]

def calculateMCSForGPRs(
    briefing: dict, isoformSeparator: str, verbose: int = 0, **kwargs
):
    """**Calculate the Minimal Cut Sets for GPRs with both AND and OR relationships**

    Given a complex GPR, we parse de GPR to a cobra model and calculate the MCSs for the model.

    :param briefing: The dictionary with the information of the GPRs.
    :param verbose: The verbosity level.

    :return: A List with the List of GPRs and a List with the MCSs.
    """
    gpr = []
    mcs = []
    parmMCS = {
        "MAX_MCS": 10000,
        "targetB": 1e-3,
        "timelimit": 60,
        "forceLength": True,
        "numWorkers": 0,
        "verbose": verbose,
        "useIndicators": True,
        "solver": "gurobi",
    }
    for key, value in kwargs.items():
        if key in parmMCS:
            parmMCS[key] = value
        else:
            warnings.warn("Parameter " + key + " is not defined in calculateMCS.")

    for i, stringGPR in enumerate(briefing["gprsANDOR"]):
        treeGenes, model = parseGPRToModel(stringGPR, i, isoformSeparator)
        if verbose > 0:
            for i in model.reactions:
                print(i)

        solutionDict = calculateMCS(
            model, MAX_LENGTH_MCS=len(treeGenes), rxnSubset=list(treeGenes), **parmMCS
        )
        for key, data in solutionDict.items():
            gpr.append(stringGPR)
            mcs.append(data["solution"])
    return gpr, mcs

@beartype
def parseGPRToModel(stringGPR: str, reactionName, isoformSeparator):
    """**Parse a GPR string to a cobra model**


    The model is used to calculate the MCS to calculate the combinations of genes that are required for the reaction.

    :param stringGPR: The GPR string to parse. (e.g. "gen1 and gen2 or gen3")
    :param reactionName: The name of the reaction to parse. Just to identify the model.

    :return:
        - **treeGenes** - A list of genes that participate in the GPR.
        - **model** - A cobra model with the GPR parsed to a metabolic model.
    """
    if isoformSeparator is not None:
        # Remove the isoform separator and the following string and numbers
        stringGPR = re.sub(f"([{isoformSeparator}][0-9]+)", "", stringGPR)
    [gprTree, treeGenes] = cobra.core.gene.parse_gpr(stringGPR)
    parentMetabolite = cobra.core.Metabolite(f"r{0}", name=f"r{0}")
    model = cobra.core.Model(f"reaction_{reactionName}")
    objectiveReaction = cobra.core.Reaction("objective")
    objectiveReaction.name = "objective"
    objectiveReaction.lower_bound = 0
    objectiveReaction.upper_bound = 1000
    objectiveReaction.add_metabolites({parentMetabolite: -1.0})
    model.add_reactions([objectiveReaction])
    model.objective = "objective"
    level = 0
    model = buildReactionFromBranch(expr=gprTree.body, model=model, level=level)

    return [treeGenes, model]

@beartype
def buildReactionFromBranch(
    expr: BoolOp,
    model: cobra.core.Model,
    level: int = 0,
    parentMetabolite: cobra.core.Metabolite = None,
) -> cobra.core.Model:
    """**Recursive function to generate metabolic model from GPR.**

    Each GPR can be expressed as a metabolic model. To that end, a boolean tree can be build, where a branch can be either an AND, an OR or a gene.
    We move through the tree, and recursively call this function on each node. Until all genes are explored (end nodes).
    As we move through the tree, we add reactions to the model. The reactions are named with a unique identifier, to avoid name conflicts.
    An example building a metabolic model from a GPR is shown below.

    .. figure:: ./figs/binaryTreeExample.png
        :height: 350px
        :width: 450 px
        :align: center

        Figure 1. Example of a binary tree.

    In the example above, the GPR is (g2 and (g5 or g6)) or g7. The function will be called recursively on each node.

    :param expr: A BoolOp object from the ast module to be explored. i.e. a tree.  (ast.BoolOp)
    :param model: A cobra model to build add the tree.
    :param level: The level of the tree, i.e. the depth of the current node.  (int)
    :param parentMetabolite: The metabolite that is the parent of the current node.  (cobra.core.Metabolite)

    :returns:
            - **model**: A cobra model with the tree added as reactions, genes have become metabolites, and the objective function set to the root node, the reaction of interest.
    """
    if isinstance(expr, BoolOp):
        branch = expr.op
        if isinstance(branch, Or):
            if parentMetabolite is None:
                parentMetabolite = cobra.core.Metabolite(
                    f"r{level}", name=f"reaction_{level}"
                )
            reaction = cobra.core.Reaction(str(uuid.uuid4()))
            reaction.name = f"OR_{level}_{parentMetabolite.id}"
            reaction.lower_bound = 0
            reaction.upper_bound = 1000
            reaction.add_metabolites({parentMetabolite: 1.0})

            for i in expr.values:
                if isinstance(i, Name):
                    level = level + 1
                    reaction = cobra.core.Reaction(str(uuid.uuid4()))
                    reaction.name = f"OR_{level}_{parentMetabolite.id}"
                    reaction.lower_bound = 0
                    reaction.upper_bound = 1000
                    reaction.add_metabolites({parentMetabolite: 1.0})
                    childMetabolite = cobra.core.Metabolite(i.id, name=i.id)
                    reaction.add_metabolites({childMetabolite: -1.0})
                    model.add_reactions([reaction])
                    if i.id in model.reactions:
                        continue
                    reaction = cobra.core.Reaction(i.id)
                    reaction.name = i.id
                    reaction.lower_bound = 0
                    reaction.upper_bound = 1000
                    reaction.add_metabolites({childMetabolite: 1.0})
                    model.add_reactions([reaction])
                else:
                    level = level + 1
                    reaction = cobra.core.Reaction(str(uuid.uuid4()))
                    reaction.name = f"OR_{level}_{parentMetabolite.id}"
                    reaction.lower_bound = 0
                    reaction.upper_bound = 1000
                    reaction.add_metabolites({parentMetabolite: 1.0})
                    childMetabolite = cobra.core.Metabolite(
                        f"branch_OR_met{level}_{parentMetabolite.id}",
                        name=f"reaction_{level}",
                    )
                    reaction.add_metabolites({childMetabolite: -1.0})
                    model.add_reactions([reaction])
                    buildReactionFromBranch(i, model, level, childMetabolite)
        elif isinstance(branch, And):
            if parentMetabolite is None:
                parentMetabolite = cobra.core.Metabolite(
                    f"r{level}", name=f"reaction_{level}"
                )
            reaction = cobra.core.Reaction(str(uuid.uuid4()))
            reaction.name = f"AND_{level}_{parentMetabolite.id}"
            reaction.lower_bound = 0
            reaction.upper_bound = 1000
            reaction.add_metabolites({parentMetabolite: 1.0})
            for i in expr.values:
                if isinstance(i, Name):
                    childMetabolite = cobra.core.Metabolite(i.id, name=i.id)
                    reaction.add_metabolites({childMetabolite: -1.0})
                    if i.id in model.reactions:
                        continue
                    newReaction = cobra.core.Reaction(i.id)
                    newReaction.name = i.id
                    newReaction.lower_bound = 0
                    newReaction.upper_bound = 1000
                    newReaction.add_metabolites({childMetabolite: 1.0})
                    model.add_reactions([newReaction])
                else:
                    level = level + 1
                    childMetabolite = cobra.core.Metabolite(
                        f"branch_AND_met{level}_{parentMetabolite.id}",
                        name=f"reaction_{level}",
                    )
                    reaction.add_metabolites({childMetabolite: -1.0})
                    buildReactionFromBranch(i, model, level, childMetabolite)
            model.add_reactions([reaction])

    return model


def mergeIsforms(isoformSeparator):
    if isoformSeparator is None:

        def __addToDict(dict: defaultdict, key: List, value) -> None:
            """**Add to dictionary**

            Internal function to add a value to a dictionary with a key. If the key does not exist, it is created,
            and the value is added to the list of values.
            If the key exists, the value is appended to the list of values.

            Dictionary mut be a `defaultdict <https://docs.python.org/3/library/collections.html#collections.defaultdict>`_.

            :param dict: Dictionary to add gene reaction interaction. (defaultdict)
            :param key: List of genes that render the reaction inactivate. (list), if is only one gene, it must be a list with one element.
            :param value: Any value can be stored in the dictionary, we store the corresponding index of the reactions perturbed. (any)

            """
            #key = frozenset(key)
            if key in dict:
                dict[key].append(value)
            else:
                dict[key] = [value]

    elif isinstance(isoformSeparator, str):

        def __addToDict(dict: defaultdict, key: List, value) -> None:
            key = [name.split(isoformSeparator)[0] for name in key]
            key = frozenset(key)
            if key in dict:
               dict[key].append(value)
            else:
                dict[key] = [value]

    else:
        raise TypeError("isoformSeparator must be a string or None")

    return __addToDict


@beartype
def buildRxnGeneMat(GPRs: List, reactions: List, genes: List):
    """**Build a reaction-gene matrix from a list of GPRs.**

    The binary matrix is build from each reaction GPR,
    each row corresponds to a reaction and each column corresponds to a gene.
    The matrix is sparse and is stored in the CSR format.

    :param GPRs: List of GPRs from a cobra model.
    :param reactions: List of reactions from a cobra model.
    :param genes: List of genes from a cobra model.

    :returns:
        - **rxnGeneMatrix:** A sparse matrix with the number of reactions in rows and the number of genes in columns. (scipy.sparse.csr_matrix)
        - **summary:** A dictionary with the number of reactions classified by the number of genes in their GPR. (dict)
    """
    rxnGeneMatrix = scipy.sparse.lil_matrix((len(reactions), len(genes)))

    # Build reaction-gene matrix
    for position, reaction in enumerate(reactions):
        GPR = reaction.gpr.genes
        for gene in GPR:
            rxnGeneMatrix[position, genes.index(gene)] = 1

    rxnGeneMatrix = rxnGeneMatrix.tocsr()

    rxnZeroGenes = np.where(rxnGeneMatrix.sum(axis=1) == 0)[0]
    rxnOneGene = np.where(rxnGeneMatrix.sum(axis=1) == 1)[0]

    # indices of reactions not in rxnZeroGenes and rxnOneGene
    indexRxnMoreThanOneGene = np.setdiff1d(
        np.arange(len(reactions)), np.concatenate((rxnZeroGenes, rxnOneGene))
    )

    # Keep only GPRs that include two or more genes
    GPRStrings = []
    for gpr in GPRs:
        if len(gpr.genes) >= 2:
            GPRStrings.append(gpr.to_string())

    # identify reactions with "or" and "and" relationships between genes
    rxnOr = [bool(re.search(" or ", i)) for i in GPRStrings]
    rxnAnd = [bool(re.search(" and ", i)) for i in GPRStrings]

    rxnOnlyOr = np.logical_and(rxnOr, np.logical_not(rxnAnd))
    rxnOnlyAnd = np.logical_and(rxnAnd, np.logical_not(rxnOr))
    rxnOrAnd = np.logical_and(rxnOr, rxnAnd)

    dictGPRstring = {
        position: reactions[position].gpr.to_string()
        for position in indexRxnMoreThanOneGene
    }

    gprOnlyOr = np.where(rxnOnlyOr)[0]
    indexRxnOnlyOr = []
    for gpr in gprOnlyOr:
        uniqueGPR = GPRStrings[gpr]
        indexRxnOnlyOr.extend(
            [key for key, value in dictGPRstring.items() if value == uniqueGPR]
        )

    gprRxnOnlyAnd = np.where(rxnOnlyAnd)[0]
    indexRxnOnlyAnd = []
    for gpr in gprRxnOnlyAnd:
        uniqueGPR = GPRStrings[gpr]
        indexRxnOnlyAnd.extend(
            [key for key, value in dictGPRstring.items() if value == uniqueGPR]
        )

    gprRxnOrAnd = np.where(rxnOrAnd)[0]
    indexRxnOrAnd = []
    for gpr in gprRxnOrAnd:
        uniqueGPR = GPRStrings[gpr]
        indexRxnOrAnd.extend(
            [key for key, value in dictGPRstring.items() if value == uniqueGPR]
        )

    # count number of reactions in each category
    nRxnZeroGenes = len(rxnZeroGenes)
    nRxnOneGene = len(rxnOneGene)
    nRxnOnlyOr = len(indexRxnOnlyOr)
    nRxnOnlyAnd = len(indexRxnOnlyAnd)
    nRxnOrAnd = len(indexRxnOrAnd)

    nRxnTotal = nRxnZeroGenes + nRxnOneGene + nRxnOnlyOr + nRxnOnlyAnd + nRxnOrAnd

    # Keep only GPRs that will be used in the analysis of minimal cut sets
    UniqueGPRStringsANDOR = [GPRStrings[i] for i in gprRxnOrAnd]

    # summarize results as a dictionary
    summary = {
        "nRxnZeroGenes": nRxnZeroGenes,
        "indexRxnZeroGenes": rxnZeroGenes,
        "nRxnOneGene": nRxnOneGene,
        "indexRxnOneGene": rxnOneGene,
        "nRxnOnlyOr": nRxnOnlyOr,
        "indexRxnOnlyOr": indexRxnOnlyOr,
        "nRxnOnlyAnd": nRxnOnlyAnd,
        "indexRxnOnlyAnd": indexRxnOnlyAnd,
        "nRxnOrAnd": nRxnOrAnd,
        "gprsANDOR": UniqueGPRStringsANDOR,
        "gprDict": dictGPRstring,
        "nRxnTotal": nRxnTotal,
    }

    return [rxnGeneMatrix, summary]


def initiateLogger(name, log_file, level=logging.INFO):
    handler = logging.FileHandler(log_file)
    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger

def transformReactionIdsIntoIndexes(gDict, reactionList):
    newGDict = defaultdict(list)
    for key, value in gDict.items():
        reactionIndexes = []
        for reaction in value:
            reactionIndexes.append(reactionList.index(reaction))
        newGDict[key] = reactionIndexes
    return newGDict



        


from .Utilities import createToyModel
from .Utilities import saveSolutions
from .Utilities import setSolver
from .Utilities import saveSolutionDict

from .ProblemInterpreters import cplexProblemInterpreter
from .ProblemInterpreters import gurobiProblemInterpreter
from .ProblemInterpreters import scipProblemInterpreter

from .ProblemDefinitions import buildDictionaryMCSProblem

from .calculateMCS import calculateMCS

from .calculateGMCS import calculateGeneMCS
from .calculateGMCS import buildGMatrix

from .Validations import checkGMCS
from .Validations import checkGMCSParallel

from .OptimizationProblem import OptimizationProblem


from distutils.core import setup
setup(
  name = 'gMCSpy',         
  packages = ['gMCSpy'],   
  version = '1.0',      
  license='MIT',        
  description = 'gMCSpy is a python package for the calculation of Genetic Minimal Cut sets (GMCS). In simple terms the idea is to take a metabolic model and calculate the genetic vulnerabilities that will render the biomass production impossible. This is done through a Mixed-Integer Linear problem (MILP) formultion and the use of a linear solver.',   
  author = 'Carlos Javier Rodriguez',                   
  author_email = 'cjrodriguezf@unav.es',      
  url = 'https://github.com/PlanesLab/gMCSpy',   
  download_url = 'https://github.com/PlanesLab/gMCSpy/archive/refs/tags/v.1.0.tar.gz',    
  keywords = ['gMCS', 'Genetic Minimal Cut Sets', 'Gurobi', 'CPLEX'],   
  install_requires=[            
          'cobra',
          'beartype',
          'scipy',
          'tqdm',
          'joblib',
          'gurobipy',
          'cplex'
          
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Science/Research',      # Define that your audience are developers
    'Topic :: Scientific/Engineering :: Bio-Informatics',
    'License :: OSI Approved :: MIT License',   
    'Programming Language :: Python :: 3.8',      
    'Programming Language :: Python :: 3.9',
    'Programming Language :: Python :: 3.10',      
    'Programming Language :: Python :: 3.11',
  ],
)